#!/usr/bin/python3

from debmutate.control import (
    add_dependency,
    parse_relations,
)

from lintian_brush.fixer import LintianIssue, control, report_result

uses_debhelper = False
misc_depends_added = []


def has_misc_depends(relations):
    for entry in parse_relations(relations):
        (head_whitespace, relation, tail_whitespace) = entry
        if any(r.name == "${misc:Depends}" for r in relation):
            return True
    return False


with control as updater:
    for entry in parse_relations(updater.source.get("Build-Depends", "")):
        (head_whitespace, relation, tail_whitespace) = entry
        if any(r.name in ("debhelper", "debhelper-compat") for r in relation):
            uses_debhelper = True
            break

    if uses_debhelper:
        for binary in updater.binaries:
            if has_misc_depends(binary.get("Depends", "")) or has_misc_depends(
                binary.get("Pre-Depends", "")
            ):
                continue
            else:
                issue = LintianIssue(
                    updater.source,
                    "debhelper-but-no-misc-depends",
                    info=binary["Package"],
                )
                if issue.should_fix():
                    binary["Depends"] = add_dependency(
                        binary.get("Depends", ""), "${misc:Depends}"
                    )
                    misc_depends_added.append(binary["Package"])
                    issue.report_fixed()


report_result(
    "Add missing ${{misc:Depends}} to Depends for {}.".format(
        ", ".join(misc_depends_added)
    )
)



#!/usr/bin/python3

import os
import re
import sys
from typing import Optional

from debmutate._rules import update_rules
from debmutate.debhelper import (
    get_debhelper_compat_level_from_control,
    read_debhelper_compat_file,
)

from lintian_brush.fixer import LintianIssue, control, report_result

file_compat_version: Optional[int]
try:
    file_compat_version = read_debhelper_compat_file("debian/compat")
except FileNotFoundError:
    file_compat_version = None


with control:
    control_compat_version = get_debhelper_compat_level_from_control(
        control.source
    )


compat_version: Optional[int]
if control_compat_version is not None and file_compat_version is not None:
    os.remove("debian/compat")
    compat_version = control_compat_version
elif control_compat_version is not None:
    compat_version = control_compat_version
elif file_compat_version is not None:
    compat_version = file_compat_version
else:
    sys.exit(0)


def drop_explicit_dh_compat(line):
    m = re.match(b"export DH_COMPAT[ \t]*=[ \t]*([0-9]+)", line)
    if m:
        rules_version = int(m.group(1).decode("utf-8"))
        if compat_version and compat_version != rules_version:
            issue = LintianIssue(
                "source",
                "declares-possibly-conflicting-debhelper-compat-versions",
                info=f"rules={rules_version} compat={compat_version}",
            )
            if issue.should_fix():
                issue.report_fixed()
                return []
        elif compat_version:
            return []
    return line


update_rules(global_line_cb=drop_explicit_dh_compat)
report_result(
    "Avoid setting debhelper compat version in debian/rules "
    "and debian/compat."
)



#!/usr/bin/python3

import os
import re
from contextlib import suppress
from functools import partial
from typing import Dict

from debmutate.copyright import CopyrightEditor

from debian.copyright import License, NotMachineReadableError
from lintian_brush.fixer import fixed_lintian_tag, report_result

SYNOPSIS_ALIAS: Dict[str, str] = {}
updated = set()


def replace_symlink_path(synopsis, m):
    path = m.group(0)
    waslink = os.path.islink(path)
    synopsis = SYNOPSIS_ALIAS.get(synopsis, synopsis)
    newpath = "/usr/share/common-licenses/" + synopsis.rstrip("+")
    if not os.path.exists(newpath) or os.path.islink(newpath):
        return path
    if not newpath.startswith(path + "-"):
        return path
    updated.add(synopsis)
    if waslink:
        fixed_lintian_tag(
            "all", "copyright-refers-to-symlink-license", info=path.lstrip("/")
        )
    fixed_lintian_tag(
        "all",
        "copyright-refers-to-versionless-license-file",
        info=path.lstrip("/"),
    )
    return newpath


with suppress(
    FileNotFoundError, NotMachineReadableError
), CopyrightEditor() as updater:
    for para in updater.copyright.all_paragraphs():
        license = para.license
        if not license or not license.text:
            continue
        changed_text = re.sub(
            "/usr/share/common-licenses/([A-Za-z0-9-.]+)",
            partial(replace_symlink_path, license.synopsis),
            license.text,
        )
        if changed_text != license.text:
            para.license = License(license.synopsis, changed_text)


report_result(
    "Refer to specific version of license {}.".format(
        ", ".join(sorted(updated))
    )
)



#!/usr/bin/python3

import re

from lintian_brush import fixable_lintian_tags

supported_tags = fixable_lintian_tags()

with open("README.md") as f:
    contents = f.read()

replacement_text = "".join([f"* {tag}\n" for tag in sorted(supported_tags)])

with open("README.md", "w") as f:
    # TODO(jelmer): Use better sentinels, just in case somebody changes
    # the current ones?
    f.write(
        re.sub(
            r"(subset of the issues:\n\n).*(\n\.\. _writing-fixers:\n)",
            "\\1" + replacement_text + "\\2",
            contents,
            flags=re.MULTILINE | re.DOTALL,
        )
    )



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.lintian_overrides."""

import os
import re
import subprocess

from breezy.tests import (
    TestCase,
    TestCaseWithTransport,
)

from lintian_brush.lintian_overrides import (
    INFO_FIXERS,
    LintianOverride,
    fix_override_info,
    get_overrides,
    load_renamed_tags,
    override_exists,
    overrides_paths,
    update_overrides_file,
)


class OverridesPathTests(TestCaseWithTransport):
    def test_no_overrides_paths(self):
        self.assertEqual([], list(overrides_paths()))

    def test_overrides_paths(self):
        self.build_tree(
            ["debian/", "debian/source/", "debian/source/lintian-overrides"]
        )
        self.assertEqual(
            ["debian/source/lintian-overrides"], list(overrides_paths())
        )


class UpdateOverridesFileTests(TestCaseWithTransport):
    def test_no_changes(self):
        CONTENT = """\
# An architecture wildcard would look like:
foo [any-i386] binary: another-tag optional-extra
"""
        self.build_tree_contents([("overrides", CONTENT)])

        def cb(lineno, override):
            return override

        self.assertFalse(update_overrides_file(cb=cb, path="overrides"))
        self.assertFileEqual(CONTENT, "overrides")

    def test_delete_last(self):
        CONTENT = """\
# An architecture wildcard would look like:
foo [any-i386] binary: another-tag optional-extra
"""
        self.build_tree_contents([("overrides", CONTENT)])

        def cb(lineno, override):
            return None

        self.assertTrue(update_overrides_file(cb=cb, path="overrides"))
        self.assertFalse(os.path.exists("overrides"))

    def test_delete(self):
        CONTENT = """\
# An architecture wildcard would look like:
foo [any-i386] binary: another-tag optional-extra
bar binary: onetag
"""
        self.build_tree_contents([("overrides", CONTENT)])

        def cb(lineno, override):
            return override if override.tag != "another-tag" else None

        self.assertTrue(update_overrides_file(cb=cb, path="overrides"))
        self.assertFileEqual("bar binary: onetag\n", "overrides")

    def test_change_set_archlist(self):
        self.build_tree_contents(
            [
                (
                    "overrides",
                    """\
# An architecture wildcard would look like:
foo binary: another-tag optional-extra
""",
                )
            ]
        )

        def cb(lineno, override):
            return LintianOverride(
                tag=override.tag,
                package=override.package,
                type=override.type,
                info=override.info,
                archlist=["any-i386"],
            )

        self.assertTrue(update_overrides_file(cb=cb, path="overrides"))
        self.assertFileEqual(
            """\
# An architecture wildcard would look like:
foo [any-i386] binary: another-tag optional-extra
""",
            "overrides",
        )


class OverrideExistsTests(TestCaseWithTransport):
    def test_override_exists(self):
        self.build_tree_contents(
            [
                ("debian/",),
                ("debian/source/",),
                (
                    "debian/source/lintian-overrides",
                    """\
blah source: patch-file-exists-but info
""",
                ),
            ]
        )
        self.assertEqual(
            [
                LintianOverride(
                    package="blah",
                    archlist=None,
                    type="source",
                    tag="patch-file-exists-but",
                    info="info",
                )
            ],
            list(get_overrides()),
        )
        self.assertTrue(override_exists("patch-file-exists-but", info="info"))
        self.assertFalse(override_exists("patch-file-exists-but", info="no"))
        self.assertTrue(
            override_exists(
                tag="patch-file-exists-but", info="info", package="blah"
            )
        )


class InfoFixerTests(TestCase):
    def test_tags_known(self):
        tags = {
            x.decode()
            for x in subprocess.check_output(
                ["lintian-explain-tags", "--list-tags"]
            ).splitlines(False)
        }
        tags.update(load_renamed_tags())
        for tag in INFO_FIXERS:
            self.assertIn(tag, tags)

    def test_valid_regexes(self):
        for v in INFO_FIXERS.values():
            if isinstance(v, tuple):
                vs = [v]
            elif isinstance(v, list):
                vs = v
            for v in vs:
                if isinstance(v, tuple):
                    try:
                        re.compile(v[0])
                    except re.error as e:
                        self.fail(f"Invalid regex {v[0]}: {e}")


INFO_FIXER_TESTS = [
    ("maintainer-manual-page", "*", "[*]"),
    ("source-is-missing", "lib/hash/md4.js", "[lib/hash/md4.js]"),
    ("source-is-missing", "lib/hash/md4.js *", "[lib/hash/md4.js]"),
    (
        "source-is-missing",
        "test/integration/client/big-simple-query-tests.js line "
        "length is 1118 characters (>512)",
        "[test/integration/client/big-simple-query-tests.js]",
    ),
    (
        "source-contains-prebuilt-javascript-object",
        "lib/hash/md5.js",
        "[lib/hash/md5.js]",
    ),
    (
        "very-long-line-length-in-source-file",
        "debian/gbp.conf *",
        "* [debian/gbp.conf:*]",
    ),
    (
        "very-long-line-length-in-source-file",
        "benchmark/samples/lorem1.txt line 3 is 881 characters long (>512)",
        "881 > 512 [benchmark/samples/lorem1.txt:3]",
    ),
    (
        "very-long-line-length-in-source-file",
        "docs/*.css line *",
        "* [docs/*.css:*]",
    ),
    (
        "missing-license-text-in-dep5-copyright",
        "debian/copyright GPL-3\\+ *",
        "GPL-3\\+ [debian/copyright:*]",
    ),
    (
        "inconsistent-appstream-metadata-license",
        "menu/peg-solitaire.appdata.xml (gpl-3.0+ != gpl-3+)",
        "menu/peg-solitaire.appdata.xml (gpl-3.0+ != gpl-3+) [debian/copyright]",
    ),
    (
        "source-ships-excluded-file",
        "compiler/gradle/wrapper/gradle-wrapper.jar",
        "compiler/gradle/wrapper/gradle-wrapper.jar [debian/copyright:*]",
    ),
    (
        "missing-license-paragraph-in-dep5-copyright",
        "artistic *",
        "artistic [debian/copyright:*]",
    ),
    (
        "script-not-executable",
        r"\[etc/lynis/plugins/*",
        r"\[etc/lynis/plugins/*",
    ),
    (
        "source-is-missing",
        "pydata_sphinx_theme/static/js/index.d8bbf5861d671d414e1a.js line length "
        "is 992 characters (>512)",
        "[pydata_sphinx_theme/static/js/index.d8bbf5861d671d414e1a.js]",
    ),
    (
        "very-long-line-length-in-source-file",
        "build/js/bootstrap-tour-standalone.js line length is 587 characters "
        "(>512)",
        "587 > 512 [build/js/bootstrap-tour-standalone.js:*]",
    ),
    ("hardening-no-relro", "usr/lib/libfoo.so", "[usr/lib/libfoo.so]"),
    ("hardening-no-pie", "usr/lib/libfoo.so", "[usr/lib/libfoo.so]"),
    (
        "jar-not-in-usr-share",
        "usr/lib/R/site-library/rJava/jri/JRI.jar",
        "[usr/lib/R/site-library/rJava/jri/JRI.jar]",
    ),
    (
        "package-installs-java-bytecode",
        "usr/lib/R/site-library/rJava/java/ArrayWrapper.class",
        "[usr/lib/R/site-library/rJava/java/ArrayWrapper.class]",
    ),
    (
        "debconf-is-not-a-registry",
        "usr/share/doc/dbconfig-common/examples/config",
        "[usr/share/doc/dbconfig-common/examples/config:*]",
    ),
    (
        "unused-debconf-template",
        "dbconfig-common/app-password-confirm",
        "dbconfig-common/app-password-confirm [*:*]",
    ),
    (
        "apache2-reverse-dependency-calls-invoke-rc.d",
        "postrm:21",
        "[postrm:21]",
    ),
    (
        "application-in-library-section",
        "libs usr/bin/cpan5.32-x86_64-linux-gnu",
        "libs [usr/bin/cpan5.32-x86_64-linux-gnu]",
    ),
    (
        "repeated-path-segment",
        "books usr/lib/acl2-8.4dfsg/books/projects/paco/books/",
        "books [usr/lib/acl2-8.4dfsg/books/projects/paco/books/]",
    ),
    ("symlink-is-self-recursive", ". usr/bin/X11", ". [usr/bin/X11]"),
    (
        "privacy-breach-google-adsense",
        "usr/share/doc/jquery-alternative-doc/index.html "
        '(choke on: "google-analytics.com/ga.js")',
        '(choke on: "google-analytics.com/ga.js") '
        "[usr/share/doc/jquery-alternative-doc/index.html]",
    ),
    (
        "systemd-service-file-refers-to-unusual-wantedby-target",
        "lib/systemd/system/autodir-home.service autodir.service",
        "autodir.service [lib/systemd/system/autodir-home.service]",
    ),
    (
        "groff-message",
        "usr/share/man/man1/connmanctl.1.gz 282: warning [p 4, 6.0i]: "
        "cannot adjust line",
        "282: warning [p 4, 6.0i]: cannot adjust line "
        "[usr/share/man/man1/connmanctl.1.gz:*]",
    ),
    ("spelling-error-in-binary", "wtH with *", "wtH with *"),
    (
        "spelling-error-in-binary",
        "usr/bin/foo wtH with",
        "wtH with [usr/bin/foo]",
    ),
    (
        "duplicate-font-file",
        "usr/lib/R/site-library/rgl/fonts/FreeMono.ttf "
        "also in fonts-freefont-ttf",
        "also in (fonts-freefont-ttf) "
        "[usr/lib/R/site-library/rgl/fonts/FreeMono.ttf]",
    ),
]


class InfoFixerDataTest(TestCase):
    def test_data(self):
        for tag, old_info, expected_info in INFO_FIXER_TESTS:
            got_info = fix_override_info(
                LintianOverride(tag=tag, info=old_info)
            )
            self.assertEqual(
                got_info,
                expected_info,
                f"Unexpected transformation for {tag}: {old_info!r} ⇒ {got_info!r} != {expected_info!r}",
            )



#!/usr/bin/python3

from contextlib import suppress

from debmutate._rules import (
    RulesEditor,
    dh_invoke_drop_with,
)

from lintian_brush.fixer import LintianIssue, report_result


def rules_find_patches_directory(makefile):
    """Find the patches directory set in debian/rules.

    Args:
        makefile: Makefile to scan
    Returns:
        path to patches directory, or None if none was found in debian/rules
    """
    try:
        val = makefile.get_variable(b"QUILT_PATCH_DIR")
    except KeyError:
        return None
    else:
        return val.decode()


def drop_quilt_with(line, target):
    newline = dh_invoke_drop_with(line, b"quilt")
    if line != newline:
        issue = LintianIssue(
            "source",
            "dh-quilt-addon-but-quilt-source-format",
            "dh ... --with quilt (line XX)",
        )
        if issue.should_fix():
            issue.report_fixed()
        else:
            newline = line
    return newline


with suppress(FileNotFoundError), open("debian/source/format") as f:
    if f.read().strip() == "3.0 (quilt)":
        with RulesEditor() as updater:
            if rules_find_patches_directory(updater.makefile) in (
                "debian/patches",
                None,
            ):
                updater.legacy_update(command_line_cb=drop_quilt_with)


report_result(
    "Don't specify --with=quilt, since package uses "
    "'3.0 (quilt)' source format."
)



#!/usr/bin/python3

import re
from typing import Optional, Set, Tuple

from debmutate._rules import Makefile, update_rules

from lintian_brush.fixer import fixed_lintian_tag, report_result

pkg_info_included = False
PATH = "/usr/share/dpkg/pkg-info.mk"
INCLUDE_LINE = b"include " + PATH.encode("ascii")

_KNOWN = [
    (b"dpkg-parsechangelog | sed -n -e 's/^Version: //p'", "DEB_VERSION"),
]

_variables: Set[str] = set()
message = None

var_sre = re.compile(b"([A-Z_]+)[ \t]*([:?]?=)[ \t]*(.*)")


def variable_defined(var):
    global _variables
    if _variables:
        return var in _variables
    with open(PATH, "rb") as f:
        for line in f:
            m = var_sre.match(line.strip())
            if m:
                _variables.add(m.group(1).decode())
    return var in _variables


def is_pkg_info_var_line(
    line: bytes,
) -> Tuple[Optional[str], bool, bool, Optional[bool]]:
    m = var_sre.match(line.strip())
    if not m:
        return (None, False, False, None)
    variable = m.group(1).decode()
    hard_assignment = m.group(2) != b"?="
    assignment = m.group(3)
    if not variable_defined(variable):
        return (variable, False, False, hard_assignment)
    m = re.match(b"\\$\\(shell (.*)\\)", assignment)
    if not m:
        return (variable, True, False, hard_assignment)
    cmd = m.group(1)
    for r, v in _KNOWN:
        if r == cmd:
            return (v, True, True, hard_assignment)
    else:
        return (variable, True, False, hard_assignment)


def drop_pkg_info_line(line):
    global pkg_info_included
    if line.strip() == INCLUDE_LINE:
        pkg_info_included = True
        return line
    (variable, variable_matches, value_matches, hard) = is_pkg_info_var_line(
        line
    )
    if not variable_matches:
        return line
    if not value_matches:
        # The assignment is different; let's include
        # the mk file *before* this assignment
        if not pkg_info_included and variable_matches:
            pkg_info_included = True
            return [INCLUDE_LINE, line]
        else:
            return line
    if hard:
        lineno = -1  # TODO(jelmer): Pass this up
        fixed_lintian_tag(
            "source",
            "debian-rules-parses-dpkg-parsechangelog",
            info=f"{variable} (line {lineno})"
        )
    if not pkg_info_included:
        pkg_info_included = True
        return INCLUDE_LINE
    return None


try:
    mf = Makefile.from_path("debian/rules")
except FileNotFoundError:
    pass
else:
    for line in mf.contents:
        if isinstance(line, bytes) and is_pkg_info_var_line(line)[1:3] == (
            True,
            True,
        ):
            update_rules(global_line_cb=drop_pkg_info_line)
            break

report_result("Avoid invoking dpkg-parsechangelog.")



#!/usr/bin/python3

import sys

from lintian_brush.fixer import LintianIssue, control, report_result, vendor
from lintian_brush.lintian import known_binary_fields, known_source_fields

# See https://people.debian.org/~mpitt/autopkgtest/README.package-tests.html
valid_field_names = set()
valid_field_names.update(known_source_fields(vendor()))
valid_field_names.update(known_binary_fields(vendor()))

case_fixed = set()


try:
    with control as updater:
        for paragraph in updater.paragraphs:
            if paragraph.get("Source"):
                para_name = "source"
            else:
                para_name = paragraph["Package"]
            for field in list(paragraph):
                if field in valid_field_names:
                    continue
                for option in valid_field_names:
                    if option.lower() != field.lower():
                        continue
                    issue = LintianIssue(
                        updater.source,
                        "cute-field",
                        f"debian/control@{para_name} {field} vs {option}",
                    )
                    if issue.should_fix():
                        issue.report_fixed()
                        value = paragraph[field]
                        del paragraph[field]
                        paragraph[option] = value
                        case_fixed.add((field, option))
                        break
except FileNotFoundError:
    sys.exit(0)

if case_fixed:
    kind = "case" + ("s" if len(case_fixed) > 1 else "")
else:
    kind = ""

fixed_str = ", ".join(
    [f"{old} ⇒ {new}" for (old, new) in sorted(list(case_fixed))]
)

report_result(f"Fix field name {kind} in debian/control ({fixed_str}).")



#!/usr/bin/python3

import sys

from debmutate.control import (
    add_dependency,
    drop_dependency,
    get_relation,
    iter_relations,
)

from lintian_brush.fixer import LintianIssue, control, report_result

added = []
removed = []
go_package = False
default_architecture = None


with control as updater:
    if any(
        iter_relations(updater.source.get("Build-Depends", ""), "golang-go")
    ):
        go_package = True
    if any(
        iter_relations(updater.source.get("Build-Depends", ""), "golang-any")
    ):
        go_package = True
    if not go_package:
        sys.exit(0)

    default_architecture = updater.source.get("Architecture")

    for binary in updater.binaries:
        if binary.get("Architecture", default_architecture) == "all":
            if "Built-Using" in binary:
                issue = LintianIssue(
                    updater.source,
                    "built-using-field-on-arch-all-package",
                    binary["Package"],
                )
                if issue.should_fix():
                    binary["Built-Using"] = drop_dependency(
                        binary["Built-Using"], "${misc:Built-Using}"
                    )
                    if not binary["Built-Using"]:
                        del binary["Built-Using"]
                    removed.append(binary["Package"])
                    issue.report_fixed()
        else:
            built_using = binary.get("Built-Using", "")
            try:
                get_relation(built_using, "${misc:Built-Using}")
            except KeyError:
                issue = LintianIssue(
                    updater.source,
                    "missing-built-using-field-for-golang-package",
                    binary["Package"],
                )
                if issue.should_fix():
                    binary["Built-Using"] = add_dependency(
                        built_using, "${misc:Built-Using}"
                    )
                    added.append(binary["Package"])
                    issue.report_fixed()

if added and removed:
    report_result(
        "Added ${{misc:Built-Using}} to {} and removed it from {}.".format(
            ", ".join(added), ", ".join(removed)
        )
    )

if added:
    report_result(
        "Add missing ${{misc:Built-Using}} to Built-Using on {}.".format(
            ", ".join(added)
        )
    )
if removed:
    report_result(
        "Remove unnecessary ${{misc:Built-Using}} for {}".format(
            ", ".join(removed)
        )
    )



#!/usr/bin/python3

from urllib.parse import urlparse

from debmutate.watch import WatchEditor

from lintian_brush.fixer import (
    LintianIssue,
    current_package_version,
    net_access_allowed,
    report_result,
    source_package_name,
)
from lintian_brush.watch import watch_entries_certainty

try:
    with WatchEditor() as updater:
        changed_entries = []
        for w in getattr(updater.watch_file, "entries", []):
            parsed_url = urlparse(w.url)
            if parsed_url.netloc != "githubredir.debian.net":
                continue
            parts = parsed_url.path.strip("/").split("/")
            if parts[0] != "github":
                # Hmm.
                continue
            issue = LintianIssue(
                "source",
                "debian-watch-file-uses-deprecated-githubredir",
                info=f"{w.url} {w.matching_pattern}",
            )
            if issue.should_fix():
                w.url = f"https://github.com/{parts[1]}/{parts[2]}/tags"
                w.matching_pattern = (
                    ".*/" + w.matching_pattern.rsplit("/", 1)[-1]
                )
                issue.report_fixed()
                changed_entries.append(w)
        if net_access_allowed():
            certainty = watch_entries_certainty(
                changed_entries,
                source_package_name(),
                expected_versions=[current_package_version().upstream_version],
                default_certainty="confident",
            )
        else:
            certainty = "confident"
except FileNotFoundError:
    pass
else:
    report_result(
        "Remove use of githubredir - see "
        "https://lists.debian.org/debian-devel-announce/2014/10/msg00000.html "
        "for details.",
        certainty=certainty,
    )



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush."""

import os
import re

from breezy.tests import (
    TestCase,
)

from debian.changelog import (
    Changelog,
)
from lintian_brush import (
    certainty_sufficient,
    certainty_to_confidence,
    min_certainty,
    version_string,
)


class LintianBrushVersion(TestCase):
    def test_matches_package_version(self):
        if not os.path.exists("debian/changelog"):
            self.skipTest(
                "no debian/changelog available. "
                "Running outside of source tree?"
            )
        with open("debian/changelog") as f:
            cl = Changelog(f, max_blocks=1)
        package_version = str(cl.version)
        m = re.match(r"^\d+\.\d+", package_version)
        assert m is not None
        package_version = m.group(0)
        self.assertEqual(package_version, version_string)


class CertaintySufficientTests(TestCase):
    def test_sufficient(self):
        self.assertTrue(certainty_sufficient("certain", "certain"))
        self.assertTrue(certainty_sufficient("certain", "possible"))
        self.assertTrue(certainty_sufficient("certain", None))
        self.assertTrue(certainty_sufficient("possible", None))
        # TODO(jelmer): Should we really always allow unknown certainties
        # through?
        self.assertTrue(certainty_sufficient(None, "certain"))  # type: ignore

    def test_insufficient(self):
        self.assertFalse(certainty_sufficient("possible", "certain"))


class CertaintyVsConfidenceTests(TestCase):
    def test_certainty_to_confidence(self):
        self.assertEqual(0, certainty_to_confidence("certain"))
        self.assertEqual(1, certainty_to_confidence("confident"))
        self.assertEqual(2, certainty_to_confidence("likely"))
        self.assertEqual(3, certainty_to_confidence("possible"))
        self.assertIs(None, certainty_to_confidence("unknown"))
        self.assertRaises(ValueError, certainty_to_confidence, "blah")


class MinimumCertaintyTests(TestCase):
    def test_minimum(self):
        self.assertEqual("certain", min_certainty([]))
        self.assertEqual("certain", min_certainty(["certain"]))
        self.assertEqual("possible", min_certainty(["possible"]))
        self.assertEqual("possible", min_certainty(["possible", "certain"]))
        self.assertEqual("likely", min_certainty(["likely", "certain"]))
        self.assertEqual(
            "possible", min_certainty(["likely", "certain", "possible"])
        )



#!/usr/bin/python3
from lintian_brush.fixer import LintianIssue, control, report_result

fields = []
packages = []


with control as updater:
    for para in updater.paragraphs:
        for k, v in para.items():
            if not v.strip():
                if para.get("Package"):
                    issue = LintianIssue(
                        updater.source,
                        "debian-control-has-empty-field",
                        info='field "{}" in package {}'.format(
                            k, para["Package"]
                        ),
                    )
                    if not issue.should_fix():
                        continue
                    issue.report_fixed()
                    packages.append(para["Package"])
                else:
                    issue = LintianIssue(
                        updater.source,
                        "debian-control-has-empty-field",
                        info=f'field "{k}" in source paragraph',
                    )
                    if not issue.should_fix():
                        continue
                    issue.report_fixed()
                fields.append(k)
                del para[k]

report_result(
    "debian/control: Remove empty control field{} {}{}.".format(
        "s" if len(fields) > 1 else "",
        ", ".join(fields),
        (" in package {}".format(", ".join(packages))) if packages else "",
    )
)



#!/usr/bin/python3

import os
import sys
from contextlib import suppress

from lintian_brush.fixer import (
    opinionated,
    report_result,
)

if not opinionated():
    sys.exit(0)

with suppress(FileNotFoundError), open("debian/patches/series") as f:
    if not f.read().strip():
        os.unlink("debian/patches/series")

report_result("Remove empty debian/patches/series.")



#!/usr/bin/python3

import os
import shlex
import sys

from debmutate._rules import Makefile, Rule, dh_invoke_get_with
from debmutate.control import (
    PkgRelation,
    add_dependency,
    is_relation_implied,
    parse_relations,
)

from lintian_brush.fixer import LintianIssue, control, report_result
from lintian_brush.lintian import LINTIAN_DATA_PATH, dh_addons, dh_commands

COMMAND_TO_DEP = {}


if not os.path.isdir(LINTIAN_DATA_PATH):
    # lintian doesn't appear to be installed
    sys.exit(2)


for command, info in dh_commands().items():
    COMMAND_TO_DEP[command] = " | ".join(info["installed_by"])

# Copied from /usr/share/lintian/lib/Lintian/Check/Debhelper.pm
COMMAND_TO_DEP.update(
    {
        "dh_apache2": "dh-apache2 | apache2-dev",
        "dh_autoreconf_clean": "dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat",
        "dh_autoreconf": "dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat",
        "dh_dkms": "dkms | dh-sequence-dkms",
        "dh_girepository": "gobject-introspection | dh-sequence-gir",
        "dh_gnome": "gnome-pkg-tools | dh-sequence-gnome",
        "dh_gnome_clean": "gnome-pkg-tools | dh-sequence-gnome",
        "dh_lv2config": "lv2core",
        "dh_make_pgxs": "postgresql-server-dev-all | postgresql-all",
        "dh_nativejava": "gcj-native-helper | default-jdk-builddep",
        "dh_pgxs_test": "postgresql-server-dev-all | postgresql-all",
        "dh_python2": "dh-python | dh-sequence-python2",
        "dh_python3": "dh-python | dh-sequence-python3",
        "dh_sphinxdoc": "sphinx | python-sphinx | python3-sphinx",
        "dh_xine": "libxine-dev | libxine2-dev",
    }
)


ADDON_TO_DEP = {}
for addon, info in dh_addons().items():
    ADDON_TO_DEP[addon] = " | ".join(info["installed_by"])


ADDON_TO_DEP.update(
    {
        # Copied from /usr/share/lintian/lib/Lintian/Check/Debhelper.pm
        "ada_library": "dh-ada-library | dh-sequence-ada-library",
        "apache2": "dh-apache2 | apache2-dev",
        "autoreconf": "dh-autoreconf | debhelper (>= 9.20160403~) | debhelper-compat",
        "cli": "cli-common-dev | dh-sequence-cli",
        "dwz": "debhelper | debhelper-compat | dh-sequence-dwz",
        "installinitramfs": "debhelper | debhelper-compat | dh-sequence-installinitramfs",
        "gnome": "gnome-pkg-tools | dh-sequence-gnome",
        "lv2config": "lv2core",
        "nodejs": "pkg-js-tools | dh-sequence-nodejs",
        "perl_dbi": "libdbi-perl | dh-sequence-perl-dbi",
        "perl_imager": "libimager-perl | dh-sequence-perl-imager",
        "pgxs": "postgresql-server-dev-all | postgresql-all",
        "pgxs_loop": "postgresql-server-dev-all | postgresql-all",
        "pypy": "dh-python | dh-sequence-pypy",
        "python2": "python2:any | python2-dev:any | dh-sequence-python2",
        "python3": "python3:any | python3-all:any | python3-dev:any | "
        "python3-all-dev:any | dh-sequence-python3",
        "scour": "scour | python-scour | dh-sequence-scour",
        "sphinxdoc": "sphinx | python-sphinx | python3-sphinx | dh-sequence-sphinxdoc",
        "systemd": "debhelper (>= 9.20160709~) | debhelper-compat | "
        "dh-sequence-systemd | dh-systemd",
        "vim_addon": "dh_vim-addon | dh-sequence-vim-addon",
    }
)

need = []

try:
    mf = Makefile.from_path("debian/rules")
except FileNotFoundError:
    sys.exit(0)

for entry in mf.contents:
    if isinstance(entry, Rule):
        for command in entry.commands():
            if command.lstrip().startswith(b"#"):
                continue
            try:
                executable = shlex.split(command.decode("utf-8", "replace"))[0]
            except (ValueError, IndexError):
                continue
            try:
                dep = COMMAND_TO_DEP[executable]
            except KeyError:
                pass
            else:
                issue = LintianIssue(
                    "source",
                    "missing-build-dependency-for-dh_-command",
                    f"{executable} => {dep}",
                )
                if not issue.should_fix():
                    continue
                need.append((dep, [issue], "command", executable))
            if executable == "dh" or executable.startswith("dh_"):
                for addon in dh_invoke_get_with(command):
                    try:
                        dep = ADDON_TO_DEP[addon]
                    except KeyError:
                        pass
                    else:
                        issue = LintianIssue(
                            "source",
                            "missing-build-dependency-for-dh-addon",
                            f"{addon} => {dep}",
                        )
                        need.append((dep, [issue], "addon", addon))


if not need:
    sys.exit(0)

changed = []

with control as updater:
    for deps, issues, kind, name in need:
        parsed = PkgRelation.parse(deps)
        is_implied = False

        if is_relation_implied(parsed, "debhelper"):
            is_implied = True

        for field in [
            "Build-Depends",
            "Build-Depends-Indep",
            "Build-Depends-Arch",
        ]:
            for _unused1, existing, _unused2 in parse_relations(
                updater.source.get(field, "")
            ):
                if is_relation_implied(parsed, existing):
                    is_implied = True

        if not is_implied:
            build_deps = updater.source.get("Build-Depends", "")
            updater.source["Build-Depends"] = add_dependency(build_deps, deps)
            for issue in issues:
                issue.report_fixed()
            changed.append((deps, issue, kind, name))

if len(changed) == 1:
    (dep, issue, kind, name) = changed[0]
    report_result(f"Add missing build dependency on {dep} for {kind} {name}.")
else:
    report_result(
        "Add missing build dependencies:"
        + "\n".join(
            f"* {dep} for {kind} {name}"
            for (dep, issue, kind, name) in changed
        )
    )



#!/usr/bin/python3
# Copyright (C) 2024 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with patches."""

__all__ = [
    "find_patches_directory",
    "tree_has_non_patches_changes",
    "read_quilt_series",
]

import os

from ._lintian_brush_rs import (
    tree_has_non_patches_changes,
)

DEFAULT_DEBIAN_PATCHES_DIR = "debian/patches"


def rules_find_patches_directory(makefile):
    """Find the patches directory set in debian/rules.

    Args:
        makefile: Makefile to scan
    Returns:
        path to patches directory, or None if none was found in debian/rules
    """
    try:
        val = makefile.get_variable(b"QUILT_PATCH_DIR")
    except KeyError:
        return None
    else:
        return val.decode()


def find_patches_directory(path):
    """Find the name of the patches directory, if any.

    Args:
      path: Root to package
    Returns:
      relative path to patches directory, or None if none exists
    """
    from debmutate._rules import Makefile

    directory = None
    try:
        mf = Makefile.from_path(os.path.join(path, "debian/rules"))
    except FileNotFoundError:
        pass
    else:
        rules_directory = rules_find_patches_directory(mf)
        if rules_directory is not None:
            directory = rules_directory
    if directory is None and os.path.exists(
        os.path.join(path, DEFAULT_DEBIAN_PATCHES_DIR)
    ):
        directory = DEFAULT_DEBIAN_PATCHES_DIR
    return directory


# TODO(jelmer): Use debmutate version
def read_quilt_series(f):
    for line in f:
        if line.startswith(b"#"):
            quoted = True
            line = line.split(b"#")[1].strip()
        else:
            quoted = False
        args = line.decode().split()
        if not args:
            continue
        patch = args[0]
        if not patch:
            continue
        options = args[1:]
        yield patch, quoted, options



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with licenses."""

import json

from . import open_binary

COMMON_LICENSES_DIR = "/usr/share/common-licenses"

FULL_LICENSE_NAME = {
    "Apache-2.0": "Apache License, Version 2.0",
    "CC0-1.0": "CC0 1.0 Universal license",
}


def load_spdx_data():
    with open_binary("spdx.json") as f:
        return json.load(f)



#!/usr/bin/python3

import sys
from email.utils import parseaddr

from debmutate.changelog import ChangelogEditor

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

TEAM_UPLOAD_LINE = "  * Team upload."
uploader_emails = []


with control as updater:
    for entry in updater.source.get("Uploaders", "").split(","):
        uploader_emails.append(parseaddr(entry)[1])


with ChangelogEditor() as updater:
    last_change = updater.changelog[0]
    if last_change.distributions != "UNRELEASED":
        sys.exit(0)
    has_team_upload = TEAM_UPLOAD_LINE in last_change.changes()
    name, email = parseaddr(last_change.author)
    if email not in uploader_emails or not has_team_upload:
        sys.exit(0)

    i = last_change._changes.index(TEAM_UPLOAD_LINE)
    del last_change._changes[i]
    if (
        i > 0
        and last_change._changes[i - 1] == ""
        and last_change._changes[i] == ""
    ):
        # Also remove the next line, if it's empty
        del last_change._changes[i]
        fixed_lintian_tag("source", "unnecessary-team-upload")

report_result("Remove unnecessary Team Upload line in changelog.")



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lintian_brush.lintian."""

from io import StringIO

from breezy.tests import TestCase

from lintian_brush.lintian import read_list_file


class ReadListFileTests(TestCase):
    def test_read_simple(self):
        f = StringIO(
            """\
# this is a comment

item1
item2

# more comments
item3
"""
        )
        self.assertEqual(
            ["item1", "item2", "item3"], list(read_list_file(f, "debian"))
        )

    def test_vendor(self):
        f = StringIO(
            """\
# this is a comment

@if-vendor-is-not ubuntu item1
@if-vendor-is-not debian item2

# more comments
item3
"""
        )
        self.assertEqual(["item1", "item3"], list(read_list_file(f, "debian")))



#!/usr/bin/python3

import os

from lintian_brush.fixer import (
    LintianIssue,
    meets_minimum_certainty,
    report_result,
)


def is_empty(path):
    verdict = "empty"
    with open(path, "rb") as f:
        for lineno, line in enumerate(f.readlines()):
            line = line.strip()
            if not line:
                continue
            if lineno == 0 and line.startswith(b"#!"):
                continue
            if line.startswith(b"#"):
                if line.lstrip(b"#") and line.strip() != b"#DEBHELPER#":
                    verdict = "some-comments"
                continue
            if line.startswith(b"set "):
                continue
            if line.startswith(b"exit "):
                continue
            return "not-empty"
    return verdict


MAINTAINER_SCRIPTS = ["prerm", "postinst", "preinst", "postrm"]
certainty = "certain"

removed = []

for entry in os.scandir("debian"):
    if entry.name in MAINTAINER_SCRIPTS:
        script = entry.name
        package = "source"
    elif "." not in entry.name:
        continue
    else:
        parts = entry.name.rsplit(".", 1)
        if len(parts) != 2:
            continue
        package = parts[0]
        script = parts[1]
        if script not in MAINTAINER_SCRIPTS:
            continue
    verdict = is_empty(entry.path)
    if verdict == "empty" or (
        verdict == "some-comments" and meets_minimum_certainty("likely")
    ):
        if verdict == "some-comments":
            certainty = "likely"
        issue = LintianIssue(package, "maintainer-script-empty", script)
        if issue.should_fix():
            removed.append((package, script))
            os.unlink(entry.path)
            issue.report_fixed()

report_result(
    "Remove empty maintainer scripts: "
    + ", ".join("{} ({})".format(*x) for x in removed),
    certainty=certainty,
)



#!/usr/bin/python3

import shlex

from debmutate._rules import update_rules

from lintian_brush.fixer import LintianIssue, report_result


def check_global(origline):
    if origline.startswith(b"export "):
        prefix = b"export "
        line = origline[len(prefix) :]
    else:
        prefix = b""
        line = origline
    try:
        (name, value) = line.split(b"=", 1)
    except ValueError:
        # Simple export of a variable, not an assignment
        return origline
    name = name.strip()
    value = value.strip()
    if name == b"DEB_LDFLAGS_MAINT_APPEND" and b"-Wl" in value:
        issue = LintianIssue(
            "source", "debian-rules-uses-as-needed-linker-flag", "line X"
        )
        if issue.should_fix():
            issue.report_fixed()
            args = shlex.split(value.decode())
            orig_args = list(args)
            for i, arg in enumerate(args):
                if arg.startswith("-Wl"):
                    ld_args = arg.split(",")
                    try:
                        ld_args.remove("--as-needed")
                    except ValueError:
                        continue
                    if not ld_args[1:]:
                        args.remove("-Wl,--as-needed")
                    else:
                        args[i] = ",".join(ld_args)
            if not args:
                return None
            if orig_args != args:
                return prefix + b"%s = %s" % (name, shlex.join(args).encode())
        return origline
    return origline


try:
    update_rules(global_line_cb=check_global, drop_related_comments=True)  # type: ignore
except TypeError:  # debmutate < 0.62
    update_rules(global_line_cb=check_global)


report_result("Avoid explicitly specifying -Wl,--as-needed linker flag.")



#!/usr/bin/python3
import os

from lintian_brush.fixer import LintianIssue, report_result

try:
    st = os.stat("debian/rules")
except FileNotFoundError:
    pass
else:
    if not (st.st_mode & 0o111):
        issue = LintianIssue("source", "debian-rules-not-executable")
        if issue.should_fix():
            os.chmod("debian/rules", 0o755)
            issue.report_fixed()


report_result("Make debian/rules executable.")



#!/usr/bin/python3

import sys
from email.utils import parseaddr

# Import convenience functions for reporting results and checking overrides
from lintian_brush.fixer import LintianIssue, control, report_result

PKG_PERL_EMAIL = "pkg-perl-maintainers@lists.alioth.debian.org"
TESTSUITE_VALUE = "autopkgtest-pkg-perl"


with control as e:
    # Parse the maintainer field and extract the email address.
    try:
        (name, email) = parseaddr(e.source["Maintainer"])
    except KeyError:
        sys.exit(2)
    if email != PKG_PERL_EMAIL:
        # Nothing to do here, it's not a pkg-perl-maintained package
        sys.exit(0)
    if e.source.get("Testsuite") == TESTSUITE_VALUE:
        sys.exit(0)
    issue = LintianIssue(
        e.source,
        "team/pkg-perl/testsuite/no-testsuite-header",
        info="autopkgtest",
    )
    if issue.should_fix():
        e.source["Testsuite"] = TESTSUITE_VALUE
        issue.report_fixed()

report_result("Set Testsuite header for perl package.", certainty="certain")



#!/usr/bin/python3

from lintian_brush.fixer import (
    LintianIssue,
    control,
    net_access_allowed,
    report_result,
)
from lintian_brush.vcs import find_secure_vcs_url

updated = set()
lp_note = False


with control as updater:
    for key, value in updater.source.items():
        if not key.startswith("Vcs-"):
            continue
        if value.startswith("lp:"):
            lp_note = True
        newvalue = find_secure_vcs_url(value, net_access=net_access_allowed())
        if newvalue is None:
            # We can't find a secure version :(
            continue
        if newvalue == value:
            # The URL was already secure
            continue
        issue = LintianIssue(
            "source",
            "vcs-field-uses-insecure-uri",
            info=f"{key} {updater.source[key]}",
        )
        if issue.should_fix():
            updater.source[key] = newvalue
            updated.add(key)
            issue.report_fixed()


if len(updated) == 1:
    description = [f"Use secure URI in Vcs control header {list(updated)[0]}."]
else:
    description = [
        "Use secure URI in Vcs control headers: {}.".format(
            ", ".join(sorted(updated))
        )
    ]
if lp_note:
    description.append("")
    description.append(
        "The lp: prefix gets expanded to http://code.launchpad.net/ for "
        "users that are not logged in on some versions of Bazaar."
    )


report_result("\n".join(description))



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.lintian_overrides import update_overrides

REMOVED_TAGS = [
    "hardening-no-stackprotector",
    "maintainer-not-full-name",
    "uploader-not-full-name",
    "uploader-address-missing",
    "no-upstream-changelog",
    "copyright-year-in-future",
    "script-calls-init-script-directly",
]

# TODO(jelmer): Check if a tag matches a binary package name.


removed = []


def fix_malformed(path, lineno, override):
    if override.tag not in REMOVED_TAGS:
        return override
    issue = LintianIssue(
        (override.type, override.package),
        "malformed-override",
        f"Unknown tag {override.tag} in line {lineno}",
    )
    if issue.should_fix():
        issue.report_fixed()
        removed.append(override.tag)
        return None
    return override


update_overrides(fix_malformed)

report_result(
    "Remove overrides for lintian tags that are no longer supported: {}".format(
        ", ".join(removed)
    )
)



#!/usr/bin/python3

import re
import sys
from typing import List

from lintian_brush.fixer import LintianIssue, report_result


def expand_tabs(line, tabwidth=8):
    ret: List[bytes] = []
    for i, _c in enumerate(line):
        if line[i : i + 1] == b"\t":
            # Round up to the next unit of tabwidth
            ret.extend([b" "] * (tabwidth - len(ret) % tabwidth))
        else:
            ret.append(line[i : i + 1])
    return b"".join(ret)


def whitespace_prefix_length(line):
    m = re.match(b"^\\s*", line)
    if not m:
        return 0
    return len(m.group(0))


def value_offset(line):
    if not line.strip():
        return None
    if line.startswith(b"#"):
        return None
    if line[0:1] in (b"\t", b" "):
        return whitespace_prefix_length(line)
    else:
        try:
            key, value = line.split(b":", 1)
        except ValueError:
            return None
        else:
            return len(key) + 1 + whitespace_prefix_length(value)


EXPECTED_HEADER = (
    b"Format: "
    b"https://www.debian.org/doc/packaging-manuals/copyright-format/1.0"
)


UNICODE_LINE_BREAK = "\u2028".encode()
UNICODE_PARAGRAPH_SEPARATOR = "\u2029".encode()


tabs_replaced = False
unicode_linebreaks_replaced = False

try:
    lines = []
    with open("debian/copyright", "rb") as f:
        line = f.readline()
        if line.rstrip().rstrip(b"/") != EXPECTED_HEADER:
            # Not a machine-readable copyright file
            sys.exit(0)
        lines.append(line)
        prev_value_offset = None
        for lineno, line in enumerate(f, start=2):
            if line.startswith(b"\t"):
                issue = LintianIssue(
                    "source",
                    "tab-in-license-text",
                    info=f"debian/copyright (paragraph at line {lineno})",
                )
                if issue.should_fix():
                    for option in [
                        b" \t" + line[1:],
                        b" \t" + line[2:],
                        b" " * 8 + line[1:],
                    ]:
                        if value_offset(option) == prev_value_offset:
                            line = option
                            break
                    else:
                        line = b" \t" + line[1:]
                    tabs_replaced = True
                    issue.report_fixed()
            if UNICODE_PARAGRAPH_SEPARATOR in line:
                # Not quite the same thing, but close enough..
                line = line.replace(
                    UNICODE_PARAGRAPH_SEPARATOR, UNICODE_LINE_BREAK * 2
                )
            if UNICODE_LINE_BREAK in line:
                unicode_linebreaks_replaced = True
                parts = line.split(UNICODE_LINE_BREAK)
                # If the line is empty, replace it with a .
                parts = [p if p else b"." for p in parts]
                parts = [parts[0]] + [b" " + p for p in parts[1:]]
                line = b"\n".join(parts)
            lines.append(line)
            prev_value_offset = value_offset(line)
except FileNotFoundError:
    pass
else:
    with open("debian/copyright", "wb") as g:
        g.writelines(lines)

sys.stdout.write("debian/copyright: ")
if tabs_replaced:
    sys.stdout.write("use spaces rather than tabs to start continuation lines")
    if unicode_linebreaks_replaced:
        sys.stdout.write(", ")
if unicode_linebreaks_replaced:
    sys.stdout.write("replace unicode linebreaks with regular linebreaks")
sys.stdout.write(".\n")
report_result()



#!/usr/bin/python3

import shlex
from contextlib import suppress

from debmutate._rules import (
    RulesEditor,
    dh_invoke_drop_argument,
    dh_invoke_drop_with,
)
from debmutate.debhelper import get_debhelper_compat_level

from lintian_brush.fixer import LintianIssue, report_result

removed_args = []
unnecessary_args = []
unnecessary_with = []


compat_version = get_debhelper_compat_level()
if compat_version and compat_version >= 10:
    unnecessary_args.append(b"--parallel")
    unnecessary_with.append(b"systemd")


def drop_unnecessary_args(line, target):
    if not line.startswith(b"dh ") and not line.startswith(b"dh_"):
        return line
    for arg in unnecessary_args:
        newline = dh_invoke_drop_argument(line, arg)
        if newline != line:
            issue = LintianIssue(
                "source",
                "debian-rules-uses-unnecessary-dh-argument",
                info=f"dh ... {arg.decode()}",
            )
            if issue.should_fix():
                removed_args.append(arg)
                line = newline
                issue.report_fixed()
    for arg in unnecessary_with:
        newline = dh_invoke_drop_with(line, arg)
        if newline != line:
            issue = LintianIssue(
                "source",
                "debian-rules-uses-unnecessary-dh-argument",
                info=f"dh ... -with={arg.decode()}",
            )
            if issue.should_fix():
                removed_args.append(b"--with=%s" % arg)
                line = newline
                issue.report_fixed()
    return line


with suppress(FileNotFoundError), RulesEditor() as updater:
    for rule in updater.makefile.iter_rules(b"%"):
        for command in rule.commands():
            if not command.startswith(b"dh"):
                continue
            argv = shlex.split(command.decode())
            if argv[0] != "dh":
                continue
            for arg in argv:
                if arg.startswith("--no-"):
                    actual = "--" + arg[len("--no-") :]
                    if actual.encode() in unnecessary_args:
                        unnecessary_args.remove(actual.encode())
                elif arg.startswith("--"):
                    actual = "--no-" + arg[len("--") :]
                    if actual.encode() in unnecessary_args:
                        unnecessary_args.remove(actual.encode())
    updater.legacy_update(drop_unnecessary_args)

report_result(
    "Drop unnecessary dh arguments: {}".format(
        ", ".join([arg.decode() for arg in removed_args])
    )
)



#!/usr/bin/python3

import re
from typing import Optional, Set, Tuple

from debmutate._rules import Makefile, update_rules

from lintian_brush.fixer import fixed_lintian_tag, opinionated, report_result

architecture_included = False
PATH = "/usr/share/dpkg/architecture.mk"
INCLUDE_LINE = b"include " + PATH.encode("ascii")

_variables: Set[str] = set()
message = None


def variable_defined(var):
    global _variables
    if _variables:
        return var in _variables
    k = {}
    with open(PATH) as f:
        for line in f:
            if line.strip().startswith("$(foreach "):
                vs = line.strip()[len("$(foreach ") :].split(",")
                k[vs[0]] = vs[1]
    for machine in k["machine"].split(" "):
        for var in k["var"].split(" "):
            _variables.add(f"DEB_{machine}_{var}")


def is_dpkg_architecture_line(
    line: bytes,
) -> Tuple[Optional[str], bool, bool, Optional[bool]]:
    m = re.match(b"([A-Z_]+)[ \t]*([:?]?=)[ \t]*(.*)", line.strip())
    if not m:
        return (None, False, False, None)
    variable = m.group(1).decode()
    hard_assignment = m.group(2) != b"?="
    assignment = m.group(3)
    if not variable_defined(variable):
        return (variable, False, False, hard_assignment)
    m = re.match(b"\\$\\(shell dpkg-architecture -q([A-Z_]+)\\)", assignment)
    if not m or variable != m.group(1).decode():
        return (variable, True, False, hard_assignment)
    else:
        return (variable, True, True, hard_assignment)


def drop_arch_line(line):
    global architecture_included
    if line.strip() == INCLUDE_LINE:
        architecture_included = True
        return line
    (
        variable,
        variable_matches,
        value_matches,
        hard,
    ) = is_dpkg_architecture_line(line)
    if not variable_matches:
        return line
    if not value_matches:
        # The assignment is different; let's include
        # the mk file *before* this assignment
        if not architecture_included and variable_matches:
            architecture_included = True
            return [INCLUDE_LINE, line]
        else:
            return line
    if hard:
        lineno = -1  # TODO(jelmer): Pass this up
        fixed_lintian_tag(
            "source",
            "debian-rules-sets-dpkg-architecture-variable",
            info=f"{variable} (line {lineno})"
        )
    if not architecture_included:
        architecture_included = True
        return INCLUDE_LINE
    return None


def update_assignment_kind(line):
    global architecture_included, message
    if line.strip() == INCLUDE_LINE:
        architecture_included = True
        return line
    (
        variable,
        variable_matches,
        value_matches,
        hard,
    ) = is_dpkg_architecture_line(line)
    if not variable_matches:
        return line
    if not value_matches:
        return line
    if not hard:
        if architecture_included:
            return None
        else:
            # Nothing wrong here
            return line
    lineno = -1  # TODO(jelmer): Pass this up
    fixed_lintian_tag(
        "source",
        "debian-rules-sets-dpkg-architecture-variable",
        info=f"{variable} (line {lineno})"
    )
    if architecture_included:
        message = "Rely on existing architecture.mk include."
        return None
    return re.sub(b"([:?]?=)", b"?=", line)


if opinionated():
    try:
        mf = Makefile.from_path("debian/rules")
    except FileNotFoundError:
        pass
    else:
        for line in mf.contents:
            if isinstance(line, bytes) and is_dpkg_architecture_line(line)[
                1:3
            ] == (True, True):
                update_rules(global_line_cb=drop_arch_line)
                message = (
                    "Rely on pre-initialized dpkg-architecture variables."
                )
                break
else:
    message = "Use ?= for assignments to architecture variables."
    update_rules(global_line_cb=update_assignment_kind)

report_result(message)



#!/usr/bin/python3

import sys

from debian.changelog import get_maintainer
from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    meets_minimum_certainty,
    report_result,
)

# TODO(jelmer): Bump this up if there's a way that we can verify that e.g. the
# ITP was filed by get_maintainer() ?
CERTAINTY = "possible"

if not meets_minimum_certainty(CERTAINTY):
    sys.exit(0)


with control as updater:
    if updater.source.get("Maintainer"):
        sys.exit(0)
    maintainer = get_maintainer()
    updater.source["Maintainer"] = "{} <{}>".format(*maintainer)
    fixed_lintian_tag(updater.source, "required-field", "Maintainer")

report_result(
    "Set the maintainer field to: {} <{}>.".format(*maintainer),
    certainty=CERTAINTY,
)



#!/usr/bin/python3

import os
import re
import sys
from contextlib import suppress

from debmutate._rules import RulesEditor, discard_pointless_override

from debian.deb822 import PkgRelation
from lintian_brush.fixer import (
    LintianIssue,
    compat_release,
    net_access_allowed,
    report_result,
)

compat_release = compat_release()


def previous_release(release):
    import distro_info

    debian = distro_info.DebianDistroInfo()
    if release in (debian.devel(), debian.testing(), "experimental"):
        return debian.stable()
    releases = debian.get_all()
    with suppress(ValueError):
        return releases[releases.index(release) - 1]
    # TODO(jelmer): Ubuntu?
    return None


VERSION_CMP_SQL = {
    "=": "=",
    ">=": ">=",
    "<=": "<=",
    ">>": ">",
    "<<": "<",
}


def package_exists(package, release, version_info):
    if not net_access_allowed():
        try:
            return package in os.environ[f"{release.upper()}_PACKAGES"].split(
                ","
            )
        except KeyError:
            return None
    try:
        from lintian_brush.udd import connect_udd_mirror
    except ModuleNotFoundError:
        return None
    with connect_udd_mirror() as udd:
        query = "SELECT True FROM packages WHERE release = %s AND package = %s"
        args = [release, package]
        if version_info is not None:
            version_cmp, version = version_info
            query += f" AND version {VERSION_CMP_SQL[version_cmp]} $3"
            args.append(version)
        with udd.cursor() as cursor:
            cursor.execute(query, tuple(args))
            row = cursor.fetchone()
            return bool(row)


def migration_done(rels):
    previous = previous_release(compat_release)
    if previous is None:
        # We can't determine if the migration is done
        return False
    for rel in rels:
        if len(rel) > 1:
            # Not sure how to handle | Replaces
            return False
        if (
            package_exists(rel[0]["name"], previous, rel[0]["version"])
            is not False
        ):
            return False
    return True


def eliminate_dbgsym_migration(line, target):
    if not line.startswith(b"dh_strip"):
        return line

    def rep_dbgsym_migration(m):
        rep = m.group(2).strip(b'"').strip(b"'").decode()
        if "$" in rep:
            # too complicated
            return m.group(0)
        rep = PkgRelation.parse_relations(rep)
        if migration_done(rep):
            issue = LintianIssue(
                "source",
                "debug-symbol-migration-possibly-complete",
                f"{m.group(0).decode().strip()} (line XX)",
            )
            if issue.should_fix():
                issue.report_fixed()
                return b""
        return m.group(0)

    line = re.sub(
        b"([ \t]+)--dbgsym-migration[= ]('[^']+'|\"[^\"]+\"|[^ ]+)",
        rep_dbgsym_migration,
        line,
    )

    if line == b"dh_strip || dh_strip":
        line = b"dh_strip"

    return line


if not os.path.exists("debian/rules"):
    sys.exit(2)
with RulesEditor() as editor:
    for rule in editor.makefile.iter_all_rules():
        newlines = []
        for line in list(rule.lines[1:]):
            if line.startswith(b"\t"):
                ret = eliminate_dbgsym_migration(line[1:], rule.target)
                newlines.append(b"\t" + ret)
            else:
                newlines.append(line)
        if rule.lines[1:] != newlines:
            rule.lines = [rule.lines[0]] + newlines
            discard_pointless_override(
                editor.makefile, rule, ignore_comments=True
            )


report_result("Drop transition for old debug package migration.")



#!/usr/bin/python3

import email.utils

from debmutate.changelog import ChangelogEditor

from lintian_brush.fixer import LintianIssue, report_result, warn

versions = []


with ChangelogEditor() as updater:
    for block in updater.changelog:
        try:
            dt = email.utils.parsedate_to_datetime(block.date)
        except (TypeError, ValueError):
            warn(f"Invalid date {block.date!r} for {block.version}")
            # parsedate_to_datetime is buggy and raises a TypeError
            # when the date is invalid.
            continue
        if dt is None:
            # Can't interpret the date. Just ignore..
            continue
        newdate = email.utils.format_datetime(dt)
        issue = LintianIssue(
            "source",
            "debian-changelog-has-wrong-day-of-week",
            info=f"{dt.day:04d}-{dt.month:02d}-{dt.year:02d} is a {dt.strftime('%A')}",
        )
        if newdate[:3] != block.date[:3] and issue.should_fix():
            block.date = newdate
            versions.append(block.version)
            issue.report_fixed()

if len(versions) == 1:
    report_result(
        "Fix day-of-week for changelog entry {}.".format(
            ", ".join([str(v) for v in versions])
        )
    )
else:
    report_result(
        "Fix day-of-week for changelog entries {}.".format(
            ", ".join([str(v) for v in versions])
        )
    )



#!/usr/bin/python3

import sys
from typing import Dict, Set

from lintian_brush.fixer import LintianIssue, control, report_result

removed: Dict[str, Set[str]] = {}

try:
    with control as updater:
        for binary in updater.binaries:
            for field, value in list(binary.items()):
                if updater.source.get(field) == value:
                    del binary[field]
                    removed.setdefault(field, set()).add(binary["Package"])
                    issue = LintianIssue(
                        updater.source,
                        "installable-field-mirrors-source",
                        info='field "{}" in package {}'.format(
                            field, binary["Package"]
                        ),
                    )
                    issue.report_fixed()
except FileNotFoundError:
    sys.exit(0)


if len(removed) == 1:
    (field, binary_packages) = list(removed.items())[0]
    report_result(
        "Remove field {} on binary package{} {} that duplicates source.".format(
            field,
            "s" if len(binary_packages) != 1 else "",
            ", ".join(sorted(binary_packages)),
        )
    )
elif len(removed) > 1:
    report_result(
        "Remove fields on binary packages that duplicate source.",
        details=[
            "Field {} from {}.".format(field, ", ".join(sorted(packages)))
            for (field, packages) in sorted(removed.items())
        ],
    )



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError
from upstream_ontologist import (
    UpstreamDatum,
)
from upstream_ontologist.guess import (
    guess_upstream_metadata_items,
)

from lintian_brush import min_certainty
from lintian_brush.fixer import (
    meets_minimum_certainty,
    report_result,
    trust_package,
)

fields = []
achieved_certainty = []


def add_upstream_metadata(copyright):
    if copyright.header.upstream_name and copyright.header.upstream_contact:
        return
    import ruamel.yaml

    try:
        with open("debian/upstream/metadata") as f:
            inp = f.read()
    except FileNotFoundError:
        upstream_metadata = {}
    else:
        yaml = ruamel.yaml.YAML(typ="safe")
        code = yaml.load(inp)
        upstream_metadata = {
            k: UpstreamDatum(k, v, certainty="certain")
            for (k, v) in code.items()
        }
    for datum in guess_upstream_metadata_items(
        ".", trust_package=trust_package()
    ):
        if not meets_minimum_certainty(datum.certainty):
            continue
        if (
            datum.field not in upstream_metadata
            or upstream_metadata[datum.field].certainty != "certain"
        ):
            upstream_metadata[datum.field] = datum

    if not copyright.header.upstream_name:
        try:
            datum = upstream_metadata["Name"]
        except KeyError:
            pass
        else:
            if datum.value:
                copyright.header.upstream_name = datum.value
                fields.append("Upstream-Name")
                achieved_certainty.append(datum.certainty)

    if not copyright.header.upstream_contact:
        try:
            datum = upstream_metadata["Contact"]
        except KeyError:
            pass
        else:
            if datum.value:
                value = datum.value
                if isinstance(value, str):
                    value = [value]
                copyright.header.upstream_contact = value
                fields.append("Upstream-Contact")
                achieved_certainty.append(datum.certainty)


with suppress(
    FileNotFoundError, NotMachineReadableError
), CopyrightEditor() as updater:
    add_upstream_metadata(updater.copyright)

certainty = min_certainty(achieved_certainty)

if len(fields) == 1:
    report_result(
        "Set field {} in debian/copyright.".format(", ".join(fields)),
        certainty=certainty,
    )
else:
    report_result(
        "Set fields {} in debian/copyright.".format(", ".join(fields)),
        certainty=certainty,
    )



#!/usr/bin/python3

import os
from contextlib import suppress

from lintian_brush.fixer import report_result

with suppress(FileNotFoundError), open("debian/source/options") as f:
    if not f.read().strip():
        os.unlink("debian/source/options")

report_result("Remove empty debian/source/options.")



#!/usr/bin/python3

from urllib.parse import urlparse

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

HOST_TO_VCS = {
    "github.com": "Git",
    "gitlab.com": "Git",
    "salsa.debian.org": "Git",
}


with control as updater:
    for field in list(updater.source):
        if not field.startswith("Vcs-") or field.lower() == "vcs-browser":
            continue
        vcs = field[4:]
        vcs_url = updater.source[field]
        host = urlparse(vcs_url).netloc.split("@")[-1]
        if host not in HOST_TO_VCS:
            continue
        actual_vcs = HOST_TO_VCS[host]
        if actual_vcs != vcs:
            fixed_lintian_tag(
                "source",
                "vcs-field-mismatch",
                f"{field} != Vcs-{actual_vcs} {vcs_url}",
            )
            del updater.source["Vcs-" + vcs]
            updater.source["Vcs-" + actual_vcs] = vcs_url
            report_result(
                f"Changed vcs type from {vcs} to {actual_vcs} based on URL."
            )



#!/usr/bin/python3

from debmutate._rules import update_rules

from lintian_brush.fixer import LintianIssue, report_result


def update_line(line, target):
    if line.strip() == b"dh_clean -k":
        issue = LintianIssue("source", "dh-clean-k-is-deprecated", "")
        if issue.should_fix():
            issue.report_fixed()
            return b"dh_prep"
    return line


update_rules(update_line)
report_result('debian/rules: Use dh_prep rather than "dh_clean -k".')



#!/usr/bin/python3
# Copyright (C) 2020 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with lintian data."""

import json
import os
from typing import Optional, TextIO

LINTIAN_DATA_PATH = "/usr/share/lintian/data"


def read_debhelper_lintian_data_file(f, sep):
    for line in f:
        if line.startswith("#"):
            continue
        if not line.strip():
            continue
        line = line.rstrip("\n")
        try:
            key, value = line.split(sep, 1)
        except ValueError:
            key = line
            value = None
        yield key, value


def read_list_file(f: TextIO, vendor: Optional[str] = None):
    for line in f:
        line = line.strip()
        if not line:
            continue
        if line.startswith("#"):
            continue
        if line.startswith("@"):
            cond, if_vendor, val = line.split(None, 2)
            if cond == "@if-vendor-is-not":
                if (
                    if_vendor
                    and vendor
                    and if_vendor.lower() == vendor.lower()
                ):
                    continue
                line = val
            else:
                raise ValueError(f"invalid check {cond!r}")
        yield line


OBSOLETE_SITES_PATH = os.path.join(
    LINTIAN_DATA_PATH, "obsolete-sites/obsolete-sites"
)
_obsolete_sites = None


def is_obsolete_site(parsed_url) -> Optional[str]:
    global _obsolete_sites
    if _obsolete_sites is None:
        with open(OBSOLETE_SITES_PATH) as f:
            _obsolete_sites = list(read_list_file(f, vendor=None))
    for site in _obsolete_sites:
        if parsed_url.hostname.endswith(site):
            return site
    else:
        return None


KNOWN_TESTS_CONTROL_FIELDS_PATH = os.path.join(
    LINTIAN_DATA_PATH, "testsuite/known-fields"
)


def _capitalize_field(field):
    return "-".join([x.capitalize() for x in field.split("-")])


def _read_test_fields(path, vendor):
    with open(path) as f:
        fields = list(read_list_file(f, vendor=vendor))

    # Older versions of lintian listed fields with all lowercase.
    if all(x == x.lower() for x in fields):
        fields = [_capitalize_field(x) for x in fields]
    return fields


def known_tests_control_fields(vendor):
    return {
        "Tests",
        "Restrictions",
        "Features",
        "Depends",
        "Tests-Directory",
        "Test-Command",
    }


KNOWN_SOURCE_FIELDS_PATH = os.path.join(
    LINTIAN_DATA_PATH, "common/source-fields"
)


def known_source_fields(vendor):
    return _read_test_fields(KNOWN_SOURCE_FIELDS_PATH, vendor)


KNOWN_BINARY_FIELDS_PATH = os.path.join(
    LINTIAN_DATA_PATH, "fields/binary-fields"
)


def known_binary_fields(vendor):
    return _read_test_fields(KNOWN_BINARY_FIELDS_PATH, vendor)


DEBHELPER_DH_COMMANDS_PATH = os.path.join(
    LINTIAN_DATA_PATH, "debhelper/dh_commands"
)
DEBHELPER_DH_COMMANDS_MANUAL_PATH = os.path.join(
    LINTIAN_DATA_PATH, "debhelper/dh_commands-manual"
)

DEBHELPER_COMMANDS_JSON_PATH = os.path.join(
    LINTIAN_DATA_PATH, "debhelper/commands.json"
)


def dh_commands():
    try:
        with open(DEBHELPER_DH_COMMANDS_PATH) as f:
            entries = set(read_debhelper_lintian_data_file(f, "="))
        with open(DEBHELPER_DH_COMMANDS_MANUAL_PATH) as f:
            entries.update(read_debhelper_lintian_data_file(f, "||"))
        return {cmd: {"installed_by": [pkg]} for (cmd, pkg) in entries}
    except FileNotFoundError:
        with open(DEBHELPER_COMMANDS_JSON_PATH) as f:
            data = json.load(f)
        return data["commands"]


def dh_addons():
    try:
        with open(os.path.join(LINTIAN_DATA_PATH, "common/dh_addons")) as f:
            return {
                addon: {"installed_by": [pkg]}
                for (addon, pkg) in read_debhelper_lintian_data_file(f, "=")
            }
    except FileNotFoundError:
        with open(
            os.path.join(LINTIAN_DATA_PATH, "debhelper/add_ons.json")
        ) as f:
            data = json.load(f)
            return data["add_ons"]



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Fix issues in XDG files."""

import os
from io import StringIO

from debmutate.reformatting import edit_formatted_file
from iniparse.ini import INIConfig, LineContainer, OptionLine


class DesktopEntryEditor:
    def __init__(self, path, allow_reformatting=False):
        self.path = path
        if allow_reformatting is None:
            allow_reformatting = (
                os.environ.get("REFORMATTING", "disallow") == "allow"
            )
        self.allow_reformatting = allow_reformatting

    def __enter__(self):
        with open(self.path) as f:
            self._orig_content = f.read()

        self._entry = INIConfig(StringIO(self._orig_content))
        self.entry = self._entry["Desktop Entry"]
        self._rewritten_content = self.dump()
        return self

    def get(self, name):
        try:
            return self.entry[name]
        except KeyError:
            return None

    def __getitem__(self, name):
        return self.entry[name]

    def __delitem__(self, name):
        del self.entry[name]

    def __setitem__(self, key, value):
        if key not in self.entry._options:
            lc = LineContainer(OptionLine(key, value, separator="="))
            self.entry._options[key] = lc
            self.entry._lines[-1].add(lc)
        else:
            self.entry._options[key].value = value

    def dump(self):
        return str(self._entry)

    def __exit__(self, exc_type, exc_val, exc_tb):
        updated_content = self.dump()

        self.changed = edit_formatted_file(
            self.path,
            self._orig_content,
            self._rewritten_content,
            updated_content,
            allow_reformatting=self.allow_reformatting,
        )
        return False



#!/usr/bin/python3

import re
import sys
from typing import List

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from lintian_brush.fixer import fixed_lintian_tag, report_result

deleted = set()
certainty = "certain"
message = "Remove listed license files (%s) from copyright."

# regex taken from /usr/share/lintian/checks/debian/copyright.pm
re_license = re.compile(r"(^|/)(COPYING[^/]*|LICENSE)$")

try:
    with CopyrightEditor() as updater:
        for paragraph in updater.copyright.all_files_paragraphs():
            files: List[str] = list()
            # access the private member because of #960278
            for f in paragraph._RestrictedWrapper__data["Files"].splitlines():
                if re_license.search(f.strip()):
                    deleted.add(f.strip())
                    fixed_lintian_tag(
                        "source",
                        "license-file-listed-in-debian-copyright",
                        info=f.strip(),
                    )
                else:
                    if files:
                        files.append(f)
                    else:
                        # First line, should not start with whitespaces.
                        files.append(f.strip())
            files_entry = "\n".join(files)
            if not files_entry.strip():
                updater.remove(paragraph)
            elif files_entry != paragraph._RestrictedWrapper__data["Files"]:
                paragraph._RestrictedWrapper__data["Files"] = files_entry

        if not deleted:
            sys.exit(0)
except (FileNotFoundError, NotMachineReadableError):
    pass
else:
    report_result(message % ", ".join(sorted(deleted)), certainty=certainty)



#!/usr/bin/python3

import os
import re
import sys

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor

if not os.path.isdir("debian/tests"):
    sys.exit(0)

for entry in os.scandir("debian/tests"):
    if not entry.is_file():
        continue
    with LineEditor(entry.path, "b") as e:
        for lineno, oldline in e:
            newline = re.sub(b"\\bADTTMP\\b", b"AUTOPKGTEST_TMP", oldline)
            if oldline == newline:
                continue
            issue = LintianIssue(
                "source",
                "uses-deprecated-adttmp",
                info=f"{entry.path} (line {lineno})"
            )
            if issue.should_fix():
                e[lineno] = newline
                issue.report_fixed()


report_result(
    "Replace use of deprecated $ADTTMP with $AUTOPKGTEST_TMP.",
    certainty="certain",
)



#!/usr/bin/python3
import os
import re
import shlex
import sys
from contextlib import suppress

from debmutate._rules import (
    Makefile,
    check_cdbs,
    dh_invoke_drop_argument,
    dh_invoke_drop_with,
    dh_invoke_replace_argument,
    update_rules,
)
from debmutate.control import (
    ensure_exact_version,
    ensure_minimum_version,
    get_relation,
)
from debmutate.debhelper import (
    MaintscriptEditor,
    MaintscriptRemoveConffile,
    read_debhelper_compat_file,
)

from lintian_brush.debhelper import (
    DEBHELPER_BUILD_STEPS,
    detect_debhelper_buildsystem,
    lowest_non_deprecated_compat_level,
    maximum_debhelper_compat_version,
)
from lintian_brush.fixer import (
    compat_release,
    control,
    current_package_version,
    fixed_lintian_tag,
    is_debcargo_package,
    report_result,
    warn,
)

new_debhelper_compat_version = maximum_debhelper_compat_version(
    compat_release()
)

uses_cdbs = check_cdbs()

if uses_cdbs:
    # cdbs doesn't appear to support debhelper 11 or 12 just yet..
    new_debhelper_compat_version = min(new_debhelper_compat_version, 10)


# If the package uses autoconf, configure doesn't contain --runstatedir, and
# specifies --without autoreconf, then we can't upgrade beyond debhelper 10. If
# we do, configure fails because debhelper >= 11 specifies --runstatedir.


# We could choose to drop the --without autoreconf, but we don't know why the
# maintainer chose to specify it.
def autoreconf_disabled():
    try:
        mf = Makefile.from_path("debian/rules")
        for line in mf.dump_lines():
            if re.findall(b"--without.*autoreconf", line):
                return True

        # Another way to disable dh_autoreconf is to add an empty override
        # rule.
        for rule in mf.iter_rules(b"override_dh_autoreconf"):
            if rule.commands():
                return False
        else:
            return True
    except FileNotFoundError:
        return False


if autoreconf_disabled():
    with suppress(IsADirectoryError, FileNotFoundError), open(
        "configure", "rb"
    ) as f:
        for line in f:
            if b"runstatedir" in line:
                break
        else:
            new_debhelper_compat_version = min(
                new_debhelper_compat_version, 10
            )
            warn(
                f"Not upgrading beyond debhelper {new_debhelper_compat_version}, since the package disables autoreconf but its configure does not provide --runstatedir."
            )


if os.path.exists("debian/compat"):
    # Package currently stores compat version in debian/compat..

    current_debhelper_compat_version = read_debhelper_compat_file(
        "debian/compat"
    )
    if current_debhelper_compat_version < new_debhelper_compat_version:
        with open("debian/compat", "w") as cf:
            cf.write(f"{new_debhelper_compat_version}\n")
    else:
        # Nothing to do
        sys.exit(2)

    with control as updater:
        updater.source["Build-Depends"] = ensure_minimum_version(
            updater.source.get("Build-Depends", ""),
            "debhelper",
            f"{new_debhelper_compat_version}~"
        )
else:
    try:
        # Assume that the compat version is set in Build-Depends
        with control as updater:
            try:
                offset, debhelper_compat_relation = get_relation(
                    updater.source.get("Build-Depends", ""), "debhelper-compat"
                )
            except KeyError:
                sys.exit(2)
            else:
                if len(debhelper_compat_relation) > 1:
                    # Not sure how to deal with this..
                    sys.exit(2)
                if debhelper_compat_relation[0].version[0] != "=":
                    # Not sure how to deal with this..
                    sys.exit(2)
                current_debhelper_compat_version = int(
                    debhelper_compat_relation[0].version[1]
                )
            if current_debhelper_compat_version < new_debhelper_compat_version:
                updater.source["Build-Depends"] = ensure_exact_version(
                    updater.source["Build-Depends"],
                    "debhelper-compat",
                    f"{new_debhelper_compat_version}"
                )
    except FileNotFoundError:
        # debcargo just uses the latest version and doesn't store debhelper
        # version explicitly.
        if is_debcargo_package():
            sys.exit(0)
        raise


# For a list of behavior changes between debhelper compat versions, see
# https://manpages.debian.org/testing/debhelper/debhelper.7.en.html#Supported_compatibility_levels


subitems = set()


def line_matches_command(target, line, command):
    if command is None:
        # Whatever
        return True

    if line.startswith(command + b" ") or line == command:
        return True

    return target == (b"override_" + command) and line.startswith(
        b"$(overridden_command)"
    )


def update_line(line, orig, new, description):
    newline = line.replace(orig, new)
    if newline != line:
        subitems.add(description)
        changed = True
    else:
        changed = False
    return newline, changed


def update_line_drop_argument(target, line, command, argument, description):
    if line_matches_command(target, line, command) and argument in line:
        line = dh_invoke_drop_argument(line, argument)
        subitems.add(description)
        return line, True
    return line, False


def update_line_replace_argument(line, old, new, description):
    newline = dh_invoke_replace_argument(line, old, new)
    if newline != line:
        subitems.add(description)
        return newline, True
    return line, False


class PybuildUpgrader:
    def __init__(self):
        # Does the dh line specify --buildsystem=pybuild?
        self.upgraded = False

    def fix_line(self, line, target):
        """Upgrade from python_distutils to pybuild."""
        line, _ = update_line(
            line,
            b"--buildsystem=python_distutils",
            b"--buildsystem=pybuild",
            "Replace python_distutils buildsystem with pybuild.",
        )
        line, _ = update_line(
            line,
            b"--buildsystem python_distutils",
            b"--buildsystem=pybuild",
            "Replace python_distutils buildsystem with pybuild.",
        )
        line, _ = update_line(
            line,
            b"-O--buildsystem=python_distutils",
            b"-O--buildsystem=pybuild",
            "Replace python_distutils buildsystem with pybuild.",
        )
        if target.decode() in DEBHELPER_BUILD_STEPS:
            step = target.decode()
        elif target.startswith(b"override_dh_auto_"):
            step = target[len(b"override_dh_auto_") :].decode()
        else:
            step = None
        if line.startswith(b"dh "):
            if b"buildsystem" not in line:
                buildsystem = detect_debhelper_buildsystem(step)
                if buildsystem == "python_distutils":
                    line += b" --buildsystem=pybuild"
                    subitems.add(
                        "Replace python_distutils buildsystem with pybuild."
                    )
                    self.upgraded = True
            else:
                if b"buildsystem=pybuild" in line:
                    self.upgraded = True
        if (
            line.startswith(b"dh_auto_")
            and b" -- " in line
            and (
                self.upgraded
                or re.match(b"--buildsystem[= ]pybuild", line)
                or detect_debhelper_buildsystem(step) == "pybuild"
            )
        ):
            line, rest = line.split(b" -- ", 1)
            if step is None:
                step = line[len(b"dh_auto_") :].split(b" ", 1)[0].decode()
            subitems.add("Replace python_distutils buildsystem with pybuild.")
            line = (
                b"PYBUILD_"
                + step.upper().encode()
                + b"_ARGS="
                + shlex.quote(rest.decode()).encode()
                + b" "
                + line
            )

        return line


def upgrade_to_dh_prep(line, target):
    """Replace 'dh_clean -k' with 'dh_prep."""
    line, changed = update_line(
        line,
        b"dh_clean -k",
        b"dh_prep",
        "debian/rules: Replace dh_clean -k with dh_prep.",
    )
    return line


class DhMissingUpgrader:
    """Replace --list-missing / --fail-missing with dh_missing."""

    def __init__(self):
        self.need_override_missing = False

    def fix_line(self, line, target):
        for arg in [b"--list-missing", b"-O--list-missing"]:
            for command in [b"dh_install", b"dh"]:
                line, changed = update_line_drop_argument(
                    target,
                    line,
                    command,
                    arg,
                    "debian/rules: Rely on default use of dh_missing rather "
                    "than using dh_install --list-missing.",
                )
        for arg in [b"--fail-missing", b"-O--fail-missing"]:
            for command in [b"dh_install", b"dh"]:
                line, changed = update_line_drop_argument(
                    target,
                    line,
                    command,
                    arg,
                    "debian/rules: Move --fail-missing argument to dh_missing.",
                )
                if changed:
                    if target == b"override_dh_install" or command == b"dh":
                        self.need_override_missing = True
                    else:
                        subitems.add(
                            "debian/rules: Move --fail-missing argument "
                            "to dh_missing."
                        )
                        return [line, b"dh_missing --fail-missing"]
        return line

    def fix_makefile(self, mf):
        if not self.need_override_missing:
            return
        try:
            [rule] = list(mf.iter_rules(b"override_dh_missing"))
        except ValueError:
            rule = mf.add_rule(b"override_dh_missing")
            rule.append_command(b"dh_missing --fail-missing")
        else:
            for i, line in enumerate(rule.lines):
                if line.startswith(b"dh_missing "):
                    if b"--fail-missing" in line:
                        return
                    else:
                        rule.lines[i] += b" --fail-missing"
            else:
                raise Exception(
                    "override_dh_missing exists, but has no call to "
                    "dh_missing"
                )


def replace_deprecated_same_arch(line, target):
    if not line.startswith(b"dh"):
        return line
    line, _ = update_line_replace_argument(
        line, b"-s", b"-a", "Replace deprecated -s with -a."
    )
    line, _ = update_line_replace_argument(
        line,
        b"--same-arch",
        b"--arch",
        "Replace deprecated --same-arch with --arch.",
    )
    return line


def upgrade_to_no_stop_on_upgrade(line, target):
    if line.startswith(b"dh ") or line.startswith(b"dh_installinit"):
        line, changed = update_line(
            line,
            b"--no-restart-on-upgrade",
            b"--no-stop-on-upgrade",
            "Replace --no-restart-on-upgrade with --no-stop-on-upgrade.",
        )
    return line


def debhelper_argument_order(line, target):
    if line.startswith(b"dh "):
        args = line.split(b" ")
        for possible_va in [b"$*", b"$@", b"${@}"]:
            try:
                x = args.index(possible_va)
            except ValueError:
                continue
            break
        else:
            return line
        val = args.pop(x)
        args.insert(1, val)
        return b" ".join(args)
    return line


def override_dh_auto_test_drop_options(rule):
    if b"override_dh_auto_test" not in rule.targets:
        return
    if (
        len(rule.lines) < 2
        or rule.lines[1] != b"ifeq (,$(filter nocheck,$(DEB_BUILD_OPTIONS)))"
        or rule.lines[-1] != b"endif"
    ):
        return
    subitems.add(
        'Drop check for DEB_BUILD_OPTIONS containing "nocheck", '
        "since debhelper now does this."
    )
    del rule.lines[1]
    del rule.lines[-1]


def upgrade_to_debhelper_12():
    pybuild_upgrader = PybuildUpgrader()
    dh_missing_upgrader = DhMissingUpgrader()
    update_rules(
        [
            debhelper_argument_order,
            replace_deprecated_same_arch,
            pybuild_upgrader.fix_line,
            upgrade_to_dh_prep,
            upgrade_to_no_stop_on_upgrade,
            dh_missing_upgrader.fix_line,
        ],
        makefile_cb=dh_missing_upgrader.fix_makefile,
    )


def upgrade_to_installsystemd(line, target):
    line = dh_invoke_drop_with(line, b"systemd")
    if line.startswith(b"dh_systemd_enable"):
        line, changed = update_line(
            line,
            b"dh_systemd_enable",
            b"dh_installsystemd",
            "Use dh_installsystemd rather than deprecated "
            "dh_systemd_enable.",
        )
    if line.startswith(b"dh_systemd_start"):
        line, changed = update_line(
            line,
            b"dh_systemd_start",
            b"dh_installsystemd",
            "Use dh_installsystemd rather than deprecated "
            "dh_systemd_start.",
        )
    return line


def rename_installsystemd_target(rule):
    rule.rename_target(
        b"override_dh_systemd_enable", b"override_dh_installsystemd"
    )
    rule.rename_target(
        b"override_dh_systemd_start", b"override_dh_installsystemd"
    )


def upgrade_to_debhelper_10():
    # dh_installinit will no longer install a file named debian/package as an
    # init script.

    with control as updater:
        for binary in updater.binaries:
            name = binary["Package"]
            if os.path.isfile(os.path.join("debian", name)):
                os.rename(
                    os.path.join("debian", name),
                    os.path.join("debian", name + ".init"),
                )
                subitems.add(f"Rename debian/{name} to debian/{name}.init.")


def upgrade_to_debhelper_11():
    update_rules(
        [upgrade_to_installsystemd], rule_cb=rename_installsystemd_target
    )
    for name in os.listdir("debian"):
        parts = name.split(".")
        if len(parts) < 2 or parts[-1] != "upstart":
            continue
        if len(parts) == 3:
            package = parts[0]
            service = parts[1]
        elif len(parts) == 2:
            package = service = parts[0]
        os.unlink(os.path.join("debian", name))
        subitems.add(f"Drop obsolete upstart file {name}.")
        with MaintscriptEditor(f"debian/{package}.maintscript") as e:
            e.append(
                MaintscriptRemoveConffile(
                    conffile=f"/etc/init/{service}.conf",
                    prior_version=current_package_version(),
                    package=None,
                )
            )


def drop_dh_missing_fail(line, target):
    """Remove dh_missing --fail-missing."""
    # Note that this is not currently replacing plain calls to dh_missing with
    # empty an "override_dh_missing" rule.

    if not line.startswith(b"dh_missing "):
        return line
    argv = line.split(b" ")
    for arg in [b"--fail-missing", b"-O--fail-missing"]:
        if arg in argv:
            # If we added the --fail-missing argument first, then
            # don't announce that here.
            try:
                subitems.remove(
                    "debian/rules: Move --fail-missing argument "
                    "to dh_missing."
                )
            except KeyError:
                subitems.add(
                    "debian/rules: Drop --fail-missing argument "
                    "to dh_missing, which is now the default."
                )
            else:
                subitems.add(
                    "debian/rules: Drop --fail-missing argument, "
                    "now the default."
                )
            argv.remove(arg)
    return b" ".join(argv)


def upgrade_to_debhelper_13():
    # dh_installtempfiles will handle d/tmpfile{,s}.  It prefer the latter
    # and warns about the former.

    with control as updater:
        for binary in updater.binaries:
            name = binary["Package"]
            tmpfile = os.path.join("debian", name + ".tmpfile")
            if os.path.isfile(tmpfile):
                os.rename(tmpfile, tmpfile + "s")
                subitems.add(f"Rename {tmpfile} to {tmpfile}s.")
        if os.path.isfile(os.path.join("debian", "tmpfile")):
            tmpfile = os.path.join("debian", "tmpfile")
            os.rename(tmpfile, tmpfile + "s")
            subitems.add(f"Rename {tmpfile} to {tmpfile}s.")

    update_rules(
        [
            # The dh_missing command will now default to --fail-missing.  This can
            # be reverted to a non-fatal warning by explicitly passing
            # --list-missing like it was in compat 12.
            drop_dh_missing_fail,
        ],
        rule_cb=override_dh_auto_test_drop_options,
    )


upgrade_to_debhelper = {
    10: upgrade_to_debhelper_10,
    11: upgrade_to_debhelper_11,
    12: upgrade_to_debhelper_12,
    13: upgrade_to_debhelper_13,
}


for version in range(
    int(str(current_debhelper_compat_version)) + 1,
    int(str(new_debhelper_compat_version)) + 1,
):
    with suppress(KeyError):
        upgrade_to_debhelper[version]()

if new_debhelper_compat_version > current_debhelper_compat_version:
    if current_debhelper_compat_version < lowest_non_deprecated_compat_level():
        kind = "deprecated"
        fixed_lintian_tag(
            "source",
            "package-uses-deprecated-debhelper-compat-version",
            info=f"{current_debhelper_compat_version}",
        )
    else:
        kind = "old"
        fixed_lintian_tag(
            "source",
            "package-uses-old-debhelper-compat-version",
            info=f"{current_debhelper_compat_version}",
        )
    description = f"Bump debhelper from {kind} {current_debhelper_compat_version} to {new_debhelper_compat_version}."

    report_result(description=description, details=sorted(subitems))



#!/usr/bin/python3

import sys
from typing import Optional

from debmutate.changelog import ChangelogEditor, Version

from lintian_brush.debbugs import find_archived_wnpp_bugs, find_wnpp_bugs
from lintian_brush.fixer import net_access_allowed, report_result

versioned_changed: Optional[Version] = None


if not net_access_allowed():
    sys.exit(0)


version_changed = None


with ChangelogEditor() as editor:
    if editor.changelog[-1].bugs_closed:
        sys.exit(0)

    wnpp_bugs = find_wnpp_bugs(editor.changelog[-1].package)
    if wnpp_bugs:
        certainty = "certain"
    else:
        wnpp_bugs = find_archived_wnpp_bugs(editor.changelog[-1].package)
        certainty = "confident"

    if not wnpp_bugs:
        sys.exit(0)

    for i, line in enumerate(editor.changelog[-1]._changes):
        if not line:
            continue

        if "Initial release" in line:
            if not line.rstrip().endswith("."):
                line = line.rstrip() + "."
            editor.changelog[-1]._changes[i] = line + (
                " Closes: #{}".format(
                    ", ".join([str(bugno) for (bugno, kind) in wnpp_bugs])
                )
            )
            version_changed = editor.changelog[-1].version
            break


if version_changed:
    report_result(
        "Add {} bugs in {}.".format(
            ", ".join(sorted({kind for (bugno, kind) in wnpp_bugs})),
            version_changed,
        ),
        certainty=certainty,
    )



#!/usr/bin/python3

from debmutate.control import (
    add_dependency,
    drop_dependency,
    get_relation,
)

from lintian_brush.fixer import control, report_result

with control as updater:
    try:
        pos, old = get_relation(
            updater.source.get("Build-Depends-Indep", ""), "debhelper-compat"
        )
    except KeyError:
        pass
    else:
        updater.source["Build-Depends"] = add_dependency(
            updater.source.get("Build-Depends", ""), old
        )
        updater.source["Build-Depends-Indep"] = drop_dependency(
            updater.source.get("Build-Depends-Indep", ""), "debhelper-compat"
        )
        if not updater.source["Build-Depends-Indep"].strip():
            del updater.source["Build-Depends-Indep"]


report_result(
    "Move debhelper-compat from Build-Depends-Indep to Build-Depends."
)



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from debian.copyright import License
from lintian_brush.fixer import LintianIssue, report_result

typos = {
    "bsd-2": "BSD-2-Clause",
    "bsd-3": "BSD-3-Clause",
    "bsd-4": "BSD-4-Clause",
    "agpl3": "AGPL-3",
    "agpl3+": "AGPL-3+",
    "lgpl2.1": "LGPL-2.1",
    "lgpl2": "LGPL-2.0",
    "lgpl3": "LGPL-3.0",
}
for i in [1, 2, 3]:
    typos[f"gplv{i}"] = f"GPL-{i}"
    typos[f"gplv{i}+"] = f"GPL-{i}+"
    typos[f"gpl{i}"] = f"GPL-{i}"
    typos[f"gpl{i}+"] = f"GPL-{i}+"

renames = {}


def fix_shortname(copyright):
    for paragraph in copyright.all_paragraphs():
        if paragraph.license is None:
            continue
        try:
            new_name = typos[paragraph.license.synopsis]
        except KeyError:
            continue
        issue = LintianIssue(
            "source",
            "invalid-short-name-in-dep5-copyright",
            info=paragraph.license.synopsis,
        )
        if issue.should_fix():
            renames[paragraph.license.synopsis] = new_name
            paragraph.license = License(new_name, paragraph.license.text)
            issue.report_fixed()


with suppress(
    FileNotFoundError, NotMachineReadableError
), CopyrightEditor() as updater:
    fix_shortname(updater.copyright)

report_result(
    "Fix invalid short license name in debian/copyright ({})".format(
        ", ".join([f"{old} ⇒ {new}" for (old, new) in renames.items()])
    )
)



#!/usr/bin/python3

import sys
from contextlib import suppress

from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    report_result,
)

# TODO(jelmer): Support unknown-priority tag

certainty = "certain"
updated = {}

with control as updater:
    if updater.source.get("Priority"):
        sys.exit(0)
    binary_priorities = set()
    for binary in updater.binaries:
        if not binary.get("Priority"):
            # TODO(jelmer): Check if all dependencies are priority standard or
            # important and if they are, consider bump the priority or not
            # setting the priority at all?
            binary["Priority"] = "optional"
            certainty = "confident"
            updated[binary["Package"]] = binary["Priority"]
            fixed_lintian_tag(binary, "recommended-field", "Priority")
        if binary.get("Priority"):
            binary_priorities.add(binary["Priority"])
    if len(binary_priorities) == 1:
        updater.source["Priority"] = binary_priorities.pop()
        for binary in updater.binaries:
            with suppress(KeyError):
                del binary["Priority"]
        report_result(
            "Set priority in source stanza, since it is the same "
            "for all packages.",
            certainty="confident",
        )
    elif updated:
        report_result(
            "Set priority for binary packages %s."
            % (["{} ({})".format(*item) for item in updated.items()])
        )



#!/usr/bin/python3

from debmutate.control import drop_dependency

from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    for field in ["Build-Depends", "Build-Depends-Indep"]:
        try:
            updater.source[field] = drop_dependency(
                updater.source[field], "build-essential"
            )
        except KeyError:
            pass
        else:
            issue = LintianIssue(
                updater.source, "build-depends-on-build-essential", field
            )
            # TODO(jelmer): Check overrides
            issue.report_fixed()


report_result("Drop unnecessary dependency on build-essential.")



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    fixed_lintian_tag,
    opinionated,
    package_is_native,
    report_result,
)

if not package_is_native():
    # Nothing to do
    sys.exit(0)

if not opinionated():
    sys.exit(0)

try:
    os.unlink("debian/upstream/metadata")
except FileNotFoundError:
    sys.exit(0)
else:
    fixed_lintian_tag("source", "upstream-metadata-in-native-source", ())

if os.listdir("debian/upstream") == []:
    os.rmdir("debian/upstream")

report_result(
    "Remove debian/upstream/metadata in native source package.",
    certainty="certain",
)



#!/usr/bin/python3
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Functions for working with watch files."""

import json
import logging
import os
import re
import urllib.error
from collections import namedtuple
from contextlib import suppress
from dataclasses import dataclass
from typing import List, Optional, Set
from urllib.parse import urlparse, urlunparse
from urllib.request import Request, urlopen

from debmutate.watch import (
    Watch,
    apply_url_mangle,
)

from . import (
    DEFAULT_URLLIB_TIMEOUT,
    USER_AGENT,
    certainty_to_confidence,
    min_certainty,
)
from .gpg import fetch_keys

# TODO(jelmer): Vary this depending on whether a new watch file was added or an
# existing one was fixed?
WATCH_FIX_VALUE = 60


@dataclass
class WatchCandidate:
    watch: Watch
    site: str
    certainty: Optional[str]
    preference: Optional[int]


class KeyRetrievalFailed(Exception):
    def __init__(self, fingerprints):
        self.fingerprints = fingerprints


COMMON_PGPSIGURL_MANGLES = [
    f"s/$/.{ext}/" for ext in ["asc", "pgp", "gpg", "sig", "sign"]
]


SignatureInfo = namedtuple("SignatureInfo", ["is_valid", "keys", "mangle"])


def probe_signature(
    r, *, pgpsigurlmangle=None, mangles=None, gpg_context=None
):
    """Try to find the signature file for a release."""
    import gpg.errors

    if gpg_context is None:
        import gpg

        gpg_context = gpg.Context(armor=True)

    if mangles is None:
        mangles = COMMON_PGPSIGURL_MANGLES

    def sig_valid(sig):
        return sig.status == 0

    if r.pgpsigurl:
        pgpsigurls = [(pgpsigurlmangle, r.pgpsigurl)]
    else:
        pgpsigurls = [
            (mangle, apply_url_mangle(mangle, r.url)) for mangle in mangles
        ]
    for mangle, pgpsigurl in pgpsigurls:
        # Try and download signatures from some predictable locations.
        try:
            resp = urlopen(pgpsigurl)
        except urllib.error.HTTPError:
            continue
        except TimeoutError:
            logging.warning("Timeout error retrieving %s", pgpsigurl)
            continue
        sig = resp.read()
        actual = urlopen(r.url).read()
        try:
            gr = gpg_context.verify(actual, sig)[1]
        except gpg.errors.GPGMEError as e:
            logging.warning(
                "Error verifying signature %s on %s: %s", pgpsigurl, r.url, e
            )
            continue
        except gpg.errors.BadSignatures as e:
            if str(e).endswith(": No public key"):
                if not fetch_keys(
                    [s.fpr for s in e.result.signatures],
                    home_dir=gpg_context.home_dir,
                ):
                    logging.warning(
                        "Unable to retrieve keys: %r", e.result.signatures
                    )
                    raise KeyRetrievalFailed(
                        [s.fpr for s in e.result.signatures]
                    ) from e
                gr = gpg_context.verify(actual, sig)[1]
            else:
                raise
        signatures = gr.signatures
        is_valid = True
        needed_keys = set()
        for sig in signatures:
            if not sig_valid(sig):
                logging.warning(
                    "Signature from %s in %s for %s not valid",
                    sig.fpr,
                    pgpsigurl,
                    r.url,
                )
                is_valid = False
            else:
                needed_keys.add(sig.fpr)
        return SignatureInfo(
            is_valid=is_valid, keys=needed_keys, mangle=mangle
        )
    else:
        return None


def candidates_from_setup_py(
    path, good_upstream_versions: Set[str], net_access=False
):
    certainty = "likely"
    # Import setuptools in case it replaces distutils
    with suppress(ImportError):
        import setuptools  # noqa: F401
    from distutils.core import run_setup

    try:
        result = run_setup(os.path.abspath(path), stop_after="config")
    except BaseException:  # noqa: PIE786
        import traceback

        traceback.print_exc()
        return
    project = result.get_name()  # type: ignore
    version = result.get_version()  # type: ignore
    if not project:
        logging.warning("no project name in setup.py")
        return
    current_version_filenames = None
    if net_access:
        # TODO(jelmer): Use ognibuild.upstream.load_pypi
        json_url = f"https://pypi.python.org/pypi/{project}/json"
        logging.info("Getting %s info on pypi (%s)", project, json_url)
        pypi_data = _load_json(json_url)
        if pypi_data is None:
            logging.warning("unable to find project %s on pypi", project)
            return
        if version in pypi_data["releases"]:
            release = pypi_data["releases"][version]
            current_version_filenames = [
                (d["filename"], d["has_sig"])
                for d in release
                if d["packagetype"] == "sdist"
            ]
    filename_regex = (
        rf"{project}-(.+)\.(?:zip|tgz|tbz|txz|(?:tar\.(?:gz|bz2|xz)))"
    )
    opts = []
    # TODO(jelmer): Set uversionmangle?
    # opts.append('uversionmangle=s/(rc|a|b|c)/~$1/')
    if current_version_filenames:
        for fn, has_sig in current_version_filenames:
            if re.match(filename_regex, fn):
                certainty = "certain"
                if has_sig:
                    opts.append("pgpsigurlmangle=s/$/.asc/")
    url = rf"https://pypi.debian.net/{project}/{filename_regex}"
    # TODO(jelmer): Add pgpsigurlmangle if has_sig==True
    w = Watch(url, opts=opts)
    yield WatchCandidate(w, "pypi", certainty=certainty, preference=1)


def find_candidates(path, good_upstream_versions, net_access=False):
    candidates = []
    if os.path.exists(os.path.join(path, "setup.py")):
        candidates.extend(
            candidates_from_setup_py(
                os.path.join(path, "setup.py"),
                good_upstream_versions,
                net_access=net_access,
            )
        )

    if os.path.exists(os.path.join(path, "debian/upstream/metadata")):
        candidates.extend(
            candidates_from_upstream_metadata(
                os.path.join(path, "debian/upstream/metadata"),
                good_upstream_versions,
                net_access=net_access,
            )
        )

    def candidate_key(candidate):
        return (
            certainty_to_confidence(candidate.certainty),
            candidate.preference,
        )

    candidates.sort(key=candidate_key)

    return candidates


def candidates_from_upstream_metadata(
    path: str, good_upstream_versions: Set[str], net_access: bool = False
):
    try:
        with open(path) as f:
            inp = f.read()
    except FileNotFoundError:
        pass
    else:
        import ruamel.yaml

        yaml = ruamel.yaml.YAML()

        yaml.preserve_quotes = True

        code = yaml.load(inp)

        for field in ["Repository", "X-Download"]:
            try:
                parsed_url = urlparse(code[field].split(" ")[0])
            except KeyError:
                pass
            else:
                if parsed_url.hostname == "github.com":
                    yield from guess_github_watch_entry(
                        parsed_url,
                        good_upstream_versions,
                        net_access=net_access,
                    )
                if parsed_url.hostname == "launchpad.net":
                    yield from guess_launchpad_watch_entry(
                        parsed_url,
                        good_upstream_versions,
                        net_access=net_access,
                    )

        archive = code.get("Archive")
        if archive == "CRAN":
            yield from guess_cran_watch_entry(code["Name"])


def guess_cran_watch_entry(name):
    w = Watch(
        rf"https://cran.r-project.org/src/contrib/{name}_([-\d.]*)\.tar\.gz"
    )
    yield WatchCandidate(w, "cran", certainty="likely", preference=0)


def guess_launchpad_watch_entry(
    parsed_url, good_upstream_versions, net_access=False
):
    if not net_access:
        return
    project = parsed_url.path.strip("/").split("/")[0]
    url = f"https://api.launchpad.net/devel/{project}/releases"
    entries = []
    while url:
        response = _load_json(url)
        entries.extend(response["entries"])
        url = response.get("next_collection_link")
    files = _load_json(entries[-1]["files_collection_link"])
    assert len(files["entries"]) > 0
    # TODO(jelmer): add
    filepattern = (
        files["entries"][0]["file_link"]
        .split("/")[-2]
        .replace(entries[-1]["version"], "(.*)")
    )
    w = Watch(
        f"https://launchpad.net/{project}/+download",
        f"https://launchpad.net/{project}/.*/{filepattern}",
    )
    yield WatchCandidate(w, "launchpad", certainty="certain", preference=0)


def guess_github_watch_entry(
    parsed_url, good_upstream_versions, net_access=False
):
    import re

    from breezy.branch import Branch

    if not net_access:
        return
    branch = Branch.open(urlunparse(parsed_url))
    tags = branch.tags.get_tag_dict()
    POSSIBLE_PATTERNS = [
        r"v(\d\S+)",
        r"(\d\S+)",
        r".*/[vV]?(\d[^\s+]+)\.tar\.gz",
    ]
    version_pattern = None
    # TODO(jelmer): Maybe use releases API instead?
    # TODO(jelmer): Automatically added mangling for
    # e.g. rc and beta
    uversionmangle: List[str] = []
    for name in sorted(tags, reverse=True):
        for pattern in POSSIBLE_PATTERNS:
            m = re.match(pattern, name)
            if not m:
                continue
            if m.group(1) in good_upstream_versions:
                version_pattern = pattern
                break
        if version_pattern:
            break
    else:
        return
    (username, project) = parsed_url.path.strip("/").split("/")
    if project.endswith(".git"):
        project = project[:-4]
    download_url = f"https://github.com/{username}/{project}/tags"
    matching_pattern = rf".*\/{version_pattern}\.tar\.gz"
    opts = [rf"filenamemangle=s/{matching_pattern}/{project}-$1\.tar\.gz/"]
    if uversionmangle:
        opts.append(r"uversionmangle=" + ";".join(uversionmangle))
    # TODO(jelmer): Check for GPG
    # opts.append(
    #    r'pgpsigurlmangle='
    #    r's/archive\/%s\.tar\.gz/releases\/download\/$1\/$1\.tar\.gz\.asc/' %
    #    version_pattern)
    w = Watch(download_url, matching_pattern, opts=opts)
    yield WatchCandidate(w, "github", certainty="certain", preference=0)


def _load_json(url):
    headers = {"User-Agent": USER_AGENT, "Accept": "application/json"}
    try:
        response = urlopen(
            Request(url, headers=headers), timeout=DEFAULT_URLLIB_TIMEOUT
        )
    except urllib.error.HTTPError as e:
        if e.status == 404:
            return None
        raise
    return json.load(response)


def candidates_from_hackage(package, good_upstream_versions, net_access=False):
    if not net_access:
        return
    url = f"https://hackage.haskell.org/package/{package}/preferred"
    versions = _load_json(url)
    if versions is None:
        return
    for version in versions["normal-version"]:
        if version in good_upstream_versions:
            break
    else:
        return
    download_url = "https://hackage.haskell.org/package/" + package
    matching_pattern = rf".*/{package}-(.*).tar.gz"
    w = Watch(download_url, matching_pattern)
    yield WatchCandidate(w, "hackage", certainty="certain", preference=1)


def fix_old_github_patterns(updater):
    ret = []
    for w in getattr(updater.watch_file, "entries", []):
        parsed_url = urlparse(w.url)

        # only applies to github.com
        if parsed_url.netloc != "github.com":
            continue

        parts = parsed_url.path.strip("/").split("/")
        if len(parts) >= 3 and parts[2] in ("tags", "releases"):  # noqa:SIM114
            pass
        elif len(parts) >= 2 and parts[0] == ".*" and parts[1] == "archive":
            pass
        else:
            continue

        parts = w.matching_pattern.split("/")
        if len(parts) > 2 and parts[-2] == "archive":
            parts.insert(-1, "refs/tags")
        w.matching_pattern = "/".join(parts)
        ret.append(w)
    return ret


def fix_github_releases(updater):
    ret = []
    for w in getattr(updater.watch_file, "entries", []):
        parsed_url = urlparse(w.url)

        # only applies to github.com
        if parsed_url.netloc != "github.com":
            continue

        parts = parsed_url.path.strip("/").split("/")
        if len(parts) >= 3 and parts[2] == "releases":
            parts[2] = "tags"
            parsed_url = parsed_url._replace(path="/".join(parts))
            w.url = parsed_url.geturl()
            ret.append(w)
    return ret


def fix_watch_issues(updater):
    ret = []
    ret.extend(fix_old_github_patterns(updater))
    ret.extend(fix_github_releases(updater))
    return ret


def watch_entries_certainty(
    entries, source_package, expected_versions=None, default_certainty="likely"
):
    certainty = "certain"
    for entry in entries:
        try:
            ret = verify_watch_entry(
                entry, source_package, expected_versions=expected_versions
            )
        except WatchEntryVerificationFailure:
            certainty = min_certainty(["possible", certainty])
        except TemporaryWatchEntryVerficationError:
            certainty = min_certainty([default_certainty, certainty])
        else:
            if not ret:
                certainty = min_certainty(["possible", certainty])
    return certainty


class WatchEntryVerificationFailure(Exception):
    """Failure verifying watch entry."""


class TemporaryWatchEntryVerficationError(Exception):
    """Temporary error verifying watch entry."""


class WatchEntryVerificationStatus:
    def __init__(self, entry, releases, missing_versions=None):
        self.entry = entry
        self.releases = {r.version: r for r in releases}
        self.missing_versions = missing_versions

    def __bool__(self):
        return not self.missing_versions and bool(self.releases)


def verify_watch_entry(
    entry: Watch,
    source_package: str,
    expected_versions: Optional[List[str]] = None,
) -> WatchEntryVerificationStatus:
    try:
        releases = list(sorted(entry.discover(source_package), reverse=True))
    except urllib.error.HTTPError as e:
        logging.warning(
            "HTTP error accessing discovery URL %s: %s.", e.geturl(), e
        )
        if (e.status or 0) // 100 == 5:
            # If the server is unhappy, then the entry could still be valid.
            raise TemporaryWatchEntryVerficationError(str(e)) from e

        raise WatchEntryVerificationFailure(str(e)) from e

    if expected_versions is None:
        return WatchEntryVerificationStatus(entry, releases=releases)

    found_versions = {str(r.version) for r in releases}
    missing_versions = set(expected_versions) - found_versions
    return WatchEntryVerificationStatus(
        entry, releases=releases, missing_versions=missing_versions
    )



#!/usr/bin/python3

import argparse
import os
import subprocess
import sys

from ruamel.yaml import YAML

from lintian_brush import fixable_lintian_tags

KNOWN_KEYS = ["tag", "status", "difficulty", "comment"]

all_tags = {
    tag.decode()
    for tag in subprocess.check_output(
        ["lintian-explain-tags", "--list-tags"]
    ).splitlines()
}

supported_tags = fixable_lintian_tags()

path = os.path.join(os.path.dirname(__file__), "tag-status.yaml")

yaml = YAML()
with open(path) as f:
    tag_status = yaml.load(f)

per_tag_status = {}
for entry in tag_status or []:
    per_tag_status[entry["tag"]] = entry
    extra_keys = set(entry.keys()) - set(KNOWN_KEYS)
    assert not extra_keys, f"Unknown keys: {extra_keys!r}"


for tag in supported_tags:
    existing = per_tag_status.get(tag)
    if existing and existing.get("status") != "implemented":
        raise Exception(
            f"tag {tag} is marked as {existing.get('status')} "
            f"in tag-status.yaml, but implemented"
        )
    per_tag_status[tag] = {"status": "implemented"}


parser = argparse.ArgumentParser()
parser.add_argument(
    "--new-tags", action="store_true", help="List missing tags."
)
parser.add_argument("--check", action="store_true", help="Check tags.")
args = parser.parse_args()

if args.new_tags:
    for tag in sorted(all_tags):
        if tag not in per_tag_status:
            print(tag)
elif args.check:
    retcode = 0
    for tag in sorted(all_tags):
        if tag not in per_tag_status:
            print(f"Missing tag: {tag}")
            retcode = 1
    sys.exit(retcode)



#!/usr/bin/python3

import logging
import os
import re
import sys
from typing import Dict, List

from debmutate.control import (
    get_relation,
    parse_standards_version,
)

from debian.changelog import Changelog
from debian.copyright import Copyright, NotMachineReadableError
from debian.deb822 import Deb822
from lintian_brush.fixer import (
    LintianIssue,
    control,
    is_debcargo_package,
    report_result,
)
from lintian_brush.standards_version import iter_standards_versions

if is_debcargo_package():
    # Not relevant
    sys.exit(0)

# For the Debian Policy upgrade checklist, see
# https://www.debian.org/doc/debian-policy/upgrading-checklist.html

# Dictionary mapping source and target versions
upgrade_path = {
    "4.1.0": "4.1.1",
    "4.1.4": "4.1.5",
    "4.2.0": "4.2.1",
    "4.3.0": "4.4.0",
    "4.4.0": "4.4.1",
    "4.4.1": "4.5.0",
    "4.5.0": "4.5.1",
    "4.5.1": "4.6.0",
    "4.6.0": "4.6.1",
    "4.6.1": "4.6.2",
}


class UpgradeCheckFailure(Exception):
    """Upgrade check failed."""

    def __init__(self, section, reason):
        self.section = section
        self.reason = reason


class UpgradeCheckUnable(Exception):
    """Unable to check upgrade requirement."""

    def __init__(self, section, reason):
        self.section = section
        self.reason = reason


def check_4_1_1():
    if not os.path.exists("debian/changelog"):
        raise UpgradeCheckFailure("4.4", "debian/changelog does not exist")
    else:
        yield "debian/changelog exists"


def check_4_4_0():
    # Check that the package uses debhelper.
    if os.path.exists("debian/compat"):
        yield "package uses debhelper"
        return
    with open("debian/control") as f:
        source = next(Deb822.iter_paragraphs(f))
        build_deps = source.get("Build-Depends", "")
        try:
            get_relation(build_deps, "debhelper-compat")
        except KeyError:
            raise UpgradeCheckFailure("4.9", "package does not use dh")
        else:
            yield "package uses debhelper"
            return


def check_4_4_1():
    # Check that there is only one Vcs field.
    vcs_fields = []
    with open("debian/control") as f:
        source = next(Deb822.iter_paragraphs(f))
        for name in source:
            if name.lower() == "vcs-browser":
                continue
            if name.lower().startswith("vcs-"):
                vcs_fields.append(name)
    if len(vcs_fields) > 1:
        raise UpgradeCheckFailure(
            "5.6.26", "package has more than one Vcs-<type> field"
        )
    elif len(vcs_fields) == 0:
        yield "package has no Vcs-<type> fields"
    else:
        yield "package has only one Vcs-<type> field"

    # Check that Files entries don't refer to directories.
    # They must be wildcards *in* the directories.
    try:
        with open("debian/copyright") as f:
            copyright = Copyright(f, strict=False)
            for para in copyright.all_files_paragraphs():
                for glob in para.files:
                    if os.path.isdir(glob):
                        raise UpgradeCheckFailure(
                            "copyright-format",
                            "Wildcards are required to match the contents of "
                            "directories",
                        )
    except FileNotFoundError:
        pass
    except NotMachineReadableError:
        pass
    else:
        yield "Files entries in debian/copyright don't refer to directories"


def check_4_1_5():
    # If epoch has changed
    with open("debian/changelog") as f:
        cl = Changelog(f, max_blocks=2)
        epochs = set()
        for block in cl:
            epochs.add(block.version.epoch)
        if len(epochs) > 1:
            # Maybe they did email debian-devel@; we don't know.
            raise UpgradeCheckUnable("5.6.12", "last release changes epoch")
    yield "Package did not recently introduce epoch"


def _poor_grep(path, needle):
    with open(path, "rb") as f:
        lines = f.readlines()
        pat = re.compile(needle)
        return any(bool(pat.search(line)) for line in lines)


def check_4_5_0():
    # Hardcoded or dynamically allocated usernames should start with an
    # underscore.

    uses_update_rc_d = False

    # For now, just check if there are any postinst / preinst script that call
    # adduser / useradd
    for entry in os.scandir("debian"):
        if not entry.name.endswith(".postinst") and not entry.name.endswith(
            ".preinst"
        ):
            continue
        if _poor_grep(entry.path, b"(adduser|useradd)"):
            raise UpgradeCheckUnable(
                "9.2.1",
                "dynamically generated usernames should start with "
                "an underscore",
            )
        if _poor_grep(entry.path, b"update-rc.d"):
            uses_update_rc_d = True
    yield "Package does not create users"

    # Including an init script is encouraged if there is no systemd unit, and
    # optional if there is (previously, it was recommended).
    for entry in os.scandir("debian"):
        if not entry.name.endswith(".init"):
            continue
        shortname = entry.name[: -len(".init")]
        if not os.path.exists(
            f"debian/{shortname}.service"
        ) and not os.path.exists(f"debian/{shortname}@.service"):
            raise UpgradeCheckFailure(
                "9.3.1",
                "packages that include system services should include "
                "systemd units",
            )
        # Use of update-rc.d is required if the package includes an init
        # script.
        if not uses_update_rc_d:
            raise UpgradeCheckFailure(
                "9.3.3",
                "update-rc usage if required if package includes init script",
            )
    if uses_update_rc_d:
        yield (
            "Package does not ship any init files without matching "
            "systemd units"
        )
        yield "Package ships init files but uses update-rc.d"
    else:
        yield "Package does not ship init files"


def check_4_5_1():
    # TODO(jelmer): check whether necessary copyright headers have been copied
    # verbatim into copyright file?

    try:
        for entry in os.scandir("debian/patches"):
            if entry.name.endswith(".series"):
                raise UpgradeCheckFailure(
                    "4.5.1", "package contains non-default series file"
                )
    except FileNotFoundError:
        yield "Package does not have any patches"
    else:
        yield "Package does not ship any non-default series files"


def check_4_2_1():
    yield from []


def check_4_6_0():
    # TODO(jelmer): No package is allowed to install files in /usr/lib64/.
    # Previously, this prohibition only applied to packages for 64-bit
    # architectures.
    for entry in os.scandir("debian"):
        if not entry.is_file():
            continue
        if _poor_grep(entry.path, b"lib64"):
            raise UpgradeCheckUnable(
                "9.1.1",
                "unable to verify whether "
                "package install files into /usr/lib/64",
            )
    else:
        yield "Package does not contain any references to lib64"


def check_4_6_1():
    # 9.1.1: Restore permission for packages for non-64-bit architectures to
    # install files to /usr/lib64/.
    # -> No need to check anything.
    yield from []


def check_4_6_2():
    # 3.8: Essential packages are only required to provide their core
    # functionality when unconfigured if they had previously been configured
    # at least once.

    # -> Loosens requirements

    # 6.5 & 6.6 The new package version is provided as an additional argument
    # following the old package version to several ``preinst``, ``prerm``,
    # and ``postrm`` maintainer script actions.

    # -> Just provides more arguments

    # 11.8.4 When computing the priority for alternatives for
    # ``/usr/bin/x-window-manager``, start with a priority of 40, not 20, and
    # don't increase the priority based on support for the (obsolete) Debian
    # menu system.
    for entry in os.scandir("debian"):
        if not entry.is_file():
            continue
        if _poor_grep(entry.path, b"x-window-manager"):
            raise UpgradeCheckUnable(
                "11.8.4",
                "unable to verify priority for "
                "/usr/bin/x-window-manager alternative",
            )
    yield "Package does not provide x-window-manager alternative"


check_requirements = {
    "4.1.1": check_4_1_1,
    "4.2.1": check_4_2_1,
    "4.4.0": check_4_4_0,
    "4.4.1": check_4_4_1,
    "4.1.5": check_4_1_5,
    "4.5.0": check_4_5_0,
    "4.5.1": check_4_5_1,
    "4.6.0": check_4_6_0,
    "4.6.1": check_4_6_1,
    "4.6.2": check_4_6_2,
}

current_version = None


verified: Dict[str, List[str]] = {}

try:
    with control as updater:
        try:
            current_version = updater.source["Standards-Version"]
        except KeyError:
            # Huh, no standards version?
            sys.exit(0)
        else:
            try:
                svs = dict(iter_standards_versions())
            except FileNotFoundError:
                dt = None
                last = None
                tag = "out-of-date-standards-version"
            else:
                last, last_dt = max(svs.items())
                try:
                    dt = svs[parse_standards_version(current_version)]
                except KeyError:
                    dt = None
                    tag = "out-of-date-standards-version"
                else:
                    age = last_dt - dt
                    if age.days > 365 * 2:
                        tag = "ancient-standards-version"
                    else:
                        tag = "out-of-date-standards-version"
            issue = LintianIssue(
                updater.source,
                tag,
                "{}{}{}".format(
                    current_version,
                    (" (released {})".format(dt.strftime("%Y-%m-%d")))
                    if dt
                    else "",
                    (
                        " (current is {})".format(
                            ".".join([str(x) for x in last])
                        )
                    )
                    if last is not None
                    else "",
                ),
            )
            if issue.should_fix():
                while current_version in upgrade_path:
                    target_version = upgrade_path[current_version]
                    check_fn = check_requirements[target_version]
                    try:
                        verified[target_version] = list(check_fn())
                    except UpgradeCheckFailure as e:
                        logging.info(
                            "Upgrade checklist validation from standards "
                            "%s ⇒ %s failed: %s: %s",
                            current_version,
                            target_version,
                            e.section,
                            e.reason,
                        )
                        break
                    except UpgradeCheckUnable as e:
                        logging.info(
                            "Unable to validate checklist from standards "
                            "%s ⇒ %s: %s: %s",
                            current_version,
                            target_version,
                            e.section,
                            e.reason,
                        )
                        break
                    current_version = target_version
                updater.source["Standards-Version"] = current_version
                issue.report_fixed()
except FileNotFoundError:
    sys.exit(0)


if current_version:
    report_result(
        f"Update standards version to {current_version}, no changes needed.",
        certainty="certain",
    )



#!/usr/bin/python3
import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    is_debcargo_package,
    meets_minimum_certainty,
    opinionated,
    package_is_native,
    report_result,
    warn,
)

if is_debcargo_package():
    sys.exit(0)

description = None

if not os.path.exists("debian/source/format"):
    orig_format = None
    format = "1.0"
    missing_source_format_issue = LintianIssue(
        "source", "missing-debian-source-format"
    )
    if not missing_source_format_issue.should_fix():
        sys.exit(0)
    missing_source_format_issue.report_fixed()
    description = "Explicitly specify source format."
else:
    with open("debian/source/format") as f:
        format = orig_format = f.read().strip()

if orig_format not in (None, "1.0"):
    sys.exit(0)

older_source_format_issue = LintianIssue(
    "source", "older-source-format", info=(orig_format or "1.0")
)

if older_source_format_issue.should_fix():
    if package_is_native():
        format = "3.0 (native)"
        description = f"Upgrade to newer source format {format}."
    else:
        from breezy import errors
        from breezy.workingtree import WorkingTree

        from lintian_brush.patches import (
            find_patches_directory,
            tree_has_non_patches_changes,
        )

        patches_directory = find_patches_directory(".")
        if patches_directory not in ("debian/patches", None):
            # Non-standard patches directory.
            warn(
                f"Tree has non-standard patches directory {patches_directory}."
            )
        else:
            try:
                tree, path = WorkingTree.open_containing(".")
            except errors.NotBranchError as e:
                if not meets_minimum_certainty("possible"):
                    warn(f"unable to open vcs to check for delta: {e}")
                    sys.exit(0)
                format = "3.0 (quilt)"
                description = f"Upgrade to newer source format {format}."
            else:
                delta = tree_has_non_patches_changes(tree, patches_directory)
                if delta:
                    warn("Tree has non-quilt changes against upstream.")
                    if opinionated():
                        format = "3.0 (quilt)"
                        description = (
                            f"Upgrade to newer source format {format}."
                        )
                        try:
                            with open("debian/source/options") as f:
                                options = list(f.readlines())
                        except FileNotFoundError:
                            options = []
                        if "single-debian-patch\n" not in options:
                            options.append("single-debian-patch\n")
                            description = description.rstrip(".") + (
                                ", enabling single-debian-patch."
                            )
                        if "auto-commit\n" not in options:
                            options.append("auto-commit\n")
                        with open("debian/source/options", "w") as f:
                            f.writelines(options)
                else:
                    format = "3.0 (quilt)"
                    description = f"Upgrade to newer source format {format}."

if not os.path.exists("debian/source"):
    os.mkdir("debian/source")

with open("debian/source/format", "w") as f:
    f.write(f"{format}\n")

if format != "1.0":
    older_source_format_issue.report_fixed()

report_result(description=description)



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from debian.copyright import License
from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.licenses import load_spdx_data

RENAMES = {
    k.lower(): v
    for k, v in {
        "Creative Commons Attribution Share-Alike (CC-BY-SA) v3.0": "CC-BY-SA-3.0",
        "Apache License Version 2.0": "Apache-2.0",
    }.items()
}


# TODO(jelmer): Ideally we'd get the list of standard SPDX
REPLACE_SPACES = {
    "public-domain",
    "mit-style",
    "bsd-style",
}

spdx_data = load_spdx_data()

RENAMES.update(
    {
        license["name"].lower(): license_id
        for license_id, license in spdx_data["licenses"].items()
    }
)
REPLACE_SPACES.update(set(spdx_data["licenses"]))
REPLACE_SPACES = {license_id.lower() for license_id in REPLACE_SPACES}
for license_id in list(REPLACE_SPACES):
    if license_id.endswith(".0"):
        REPLACE_SPACES.add(license_id[:-2])


def fix_spaces(copyright):
    for paragraph in copyright.all_paragraphs():
        if not paragraph.license:
            continue
        if " " not in paragraph.license.synopsis:
            continue
        ors = paragraph.license.synopsis.replace(" | ", " or ").split(" or ")
        names = []
        for name in ors:
            if name.lower() in RENAMES:
                name = RENAMES[name.lower()]
            elif name.replace(" ", "-").lower() in REPLACE_SPACES:
                name = name.replace(" ", "-")
            names.append(name)
        newsynopsis = " or ".join(names)
        if newsynopsis != paragraph.license.synopsis:
            fixed_lintian_tag(
                "source",
                "space-in-std-shortname-in-dep5-copyright",
                f"{paragraph.license.synopsis} (line XX)",
            )
            paragraph.license = License(newsynopsis, paragraph.license.text)


with suppress(
    FileNotFoundError, NotMachineReadableError
), CopyrightEditor() as updater:
    fix_spaces(updater.copyright)

report_result("Replace spaces in short license names with dashes.")



#!/usr/bin/python3

import os
from contextlib import suppress

from lintian_brush.fixer import LintianIssue, report_result

with suppress(FileNotFoundError, NotADirectoryError), open(
    "debian/tests/control"
) as f:
    if f.read().strip() == "":
        issue = LintianIssue("source", "empty-debian-tests-control")
        if issue.should_fix():
            os.unlink("debian/tests/control")
            if not os.listdir("debian/tests"):
                os.rmdir("debian/tests")
            issue.report_fixed()

report_result("Remove empty debian/tests/control.")



#!/usr/bin/python3

from contextlib import suppress

from debmutate.watch import WatchEditor

from lintian_brush.fixer import (
    current_package_version,
    net_access_allowed,
    report_result,
    source_package_name,
)
from lintian_brush.watch import (
    fix_old_github_patterns,
    watch_entries_certainty,
)

certainty = "certain"


with suppress(FileNotFoundError), WatchEditor() as updater:
    changed_entries = fix_old_github_patterns(updater)
    if net_access_allowed():
        certainty = watch_entries_certainty(
            changed_entries,
            source_package_name(),
            expected_versions=[current_package_version().upstream_version],
        )
    else:
        certainty = "likely"


report_result(
    "Update pattern for GitHub archive URLs from /<org>/<repo>/tags page"
    "/<org>/<repo>/archive/<tag> → /<org>/<repo>/archive/refs/tags/<tag>.",
    certainty=certainty,
)



#!/usr/bin/python3

import sys
from contextlib import suppress

from lintian_brush.fixer import (
    LintianIssue,
    control,
    report_result,
)
from lintian_brush.section import (
    find_expected_section,
    get_name_section_mappings,
)

binary_sections_set = set()
source_section_set = False
regexes = None

with control as updater:
    if updater.source.get("Section"):
        sys.exit(0)
    binary_sections = set()
    for binary in updater.binaries:
        if not binary.get("Section"):
            if regexes is None:
                regexes = get_name_section_mappings()
            section = find_expected_section(regexes, binary["Package"])
            issue = LintianIssue(binary, "recommended-field", "Section")
            if section and issue.should_fix():
                binary["Section"] = section
                binary_sections_set.add(binary["Package"])
                issue.report_fixed()
        if binary.get("Section"):
            binary_sections.add(binary["Section"])
    issue = LintianIssue(updater.source, "recommended-field", "Section")
    if len(binary_sections) == 1 and issue.should_fix():
        updater.source["Section"] = binary_sections.pop()
        for binary in updater.binaries:
            with suppress(KeyError):
                del binary["Section"]
        issue.report_fixed()
        source_section_set = True
    if source_section_set and binary_sections_set:
        report_result(
            "Section field set in source based on binary package names.",
            certainty="certain",
        )
    elif source_section_set:
        report_result(
            "Section field set in source stanza rather than binary packages.",
            certainty="certain",
        )
    elif binary_sections_set:
        report_result(
            "Section field set for binary packages {} based on name.".format(
                ", ".join(sorted(binary_sections_set))
            ),
            certainty="certain",
        )



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    opinionated,
    report_result,
)
from lintian_brush.patches import read_quilt_series

if not opinionated():
    # In a lot of cases, it seems like removing the patch is not the right
    # thing to do.
    sys.exit(0)

try:
    patches = set()
    with open("debian/patches/series", "rb") as f:
        for patch, _quoted, _options in read_quilt_series(f):
            patches.add(patch)
except FileNotFoundError:
    sys.exit(0)

removed = set()

for patch in os.listdir("debian/patches"):
    path = os.path.join("debian", "patches", patch)
    # Don't delete the series file
    if patch in ("series", "00list"):
        continue
    # Ignore everything that is not a regular file
    if not os.path.isfile(path):
        continue
    # Ignore any README files
    if patch.startswith("README"):
        continue
    # Ignore any notes files
    if patch.startswith("notes"):
        continue
    # Ignore everything that is listed in series
    if patch in patches:
        continue
    issue = LintianIssue(
        "source", "patch-file-present-but-not-mentioned-in-series", patch
    )
    if not issue.should_fix():
        continue
    removed.add(patch)
    os.unlink(path)
    issue.report_fixed()


description = (
    "Remove patch{} {} that {} missing from debian/patches/series.".format(
        "es" if len(removed) > 1 else "",
        ", ".join(sorted(removed)),
        "is" if len(removed) == 1 else "are",
    )
)
report_result(description)



#!/usr/bin/python3

import os
import sys

from debmutate.control import delete_from_list
from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.lintian import LINTIAN_DATA_PATH

removed_restrictions = []


if not os.path.exists("debian/tests/control"):
    sys.exit(0)

KNOWN_OBSOLETE_RESTRICTIONS_PATH = os.path.join(
    LINTIAN_DATA_PATH, "testsuite/known-obsolete-restrictions"
)
DEPRECATED_RESTRICTIONS = []

try:
    with open(KNOWN_OBSOLETE_RESTRICTIONS_PATH) as f:
        for line in f:
            if line.startswith("#"):
                continue
            if not line.strip():
                continue
            DEPRECATED_RESTRICTIONS.append(line.strip())
except FileNotFoundError:
    sys.exit(2)

with Deb822Editor("debian/tests/control") as updater:
    for paragraph in updater.paragraphs:
        restrictions = paragraph.get("Restrictions", "").split(",")
        if restrictions == [""]:
            continue
        to_delete = []
        for restriction in list(restrictions):
            if restriction.strip() in DEPRECATED_RESTRICTIONS:
                to_delete.append(restriction.strip())
                fixed_lintian_tag(
                    "source",
                    "obsolete-runtime-tests-restriction",
                    f"{restriction.strip()} in line XX",
                )
        if to_delete:
            removed_restrictions.extend(to_delete)
            paragraph["Restrictions"] = delete_from_list(
                paragraph["Restrictions"], to_delete
            )
            if not paragraph["Restrictions"].strip():
                del paragraph["Restrictions"]


if "needs-recommends" in removed_restrictions:
    # This is Certainty: possible, since the package may actually rely on the
    # (direct? indirect?) recommends, in which case we'd want to add them to
    # Depends.
    certainty = "possible"
else:
    certainty = "certain"


report_result(
    "Drop deprecated restriction{} {}. See "
    "https://salsa.debian.org/ci-team/autopkgtest/tree/"
    "master/doc/README.package-tests.rst".format(
        "s" if len(removed_restrictions) > 1 else "",
        ", ".join(removed_restrictions),
    ),
    certainty=certainty,
)



#!/usr/bin/python3

import os
import re

from debmutate.reformatting import GeneratedFile, check_generated_file

from lintian_brush.fixer import LintianIssue, report_result

trailing_whitespace_re = re.compile(b"\\s*\n$")
trailing_space_re = re.compile(b"[ ]*\n$")


def strip_whitespace(line: bytes, strip_tabs=True) -> bytes:
    if strip_tabs:
        pat = trailing_whitespace_re
    else:
        pat = trailing_space_re
    return re.sub(pat, b"\n", line)


def file_strip_whitespace(
    path: str,
    strip_tabs: bool = True,
    strip_trailing_empty_lines: bool = True,
    delete_new_empty_line: bool = False,
) -> bool:
    newlines = []
    changed = False
    try:
        with open(path, "rb") as f:
            for lineno, line in enumerate(f, 1):
                newline = strip_whitespace(line, strip_tabs=strip_tabs)
                if newline != line:
                    issue = LintianIssue(
                        "source",
                        "trailing-whitespace",
                        info=f"{path} (line {lineno})"
                    )
                    if issue.should_fix():
                        issue.report_fixed()
                        changed = True
                        if newline == b"\n" and delete_new_empty_line:
                            continue
                newlines.append(newline)
    except FileNotFoundError:
        return False
    if strip_trailing_empty_lines:
        while newlines and newlines[-1] == b"\n":
            issue = LintianIssue(
                "source",
                "trailing-whitespace",
                info=f"{path} (line {len(newlines) - 1})"
            )
            if issue.should_fix():
                issue.report_fixed()
                changed = True
                newlines.pop(-1)
            else:
                break
    if changed:
        with open(path, "wb") as f:
            f.writelines(newlines)
        return True
    return False


file_strip_whitespace("debian/changelog", strip_tabs=True)
file_strip_whitespace("debian/rules", strip_tabs=False)

try:
    check_generated_file("debian/control")
except GeneratedFile:
    changed = False
    for entry in os.scandir("debian"):
        if not entry.name.startswith("control.") or entry.name.endswith("~"):
            continue
        if entry.name.endswith(".m4"):
            continue
        if file_strip_whitespace(
            entry.path, strip_tabs=True, delete_new_empty_line=True
        ):
            changed = True
    if changed:
        file_strip_whitespace(
            "debian/control", strip_tabs=True, delete_new_empty_line=True
        )
else:
    file_strip_whitespace(
        "debian/control", strip_tabs=True, delete_new_empty_line=True
    )

report_result("Trim trailing whitespace.")



#!/usr/bin/python3

import os
import re

from debmutate.control import ensure_some_version

from lintian_brush.fixer import (
    LintianIssue,
    control,
    report_result,
)
from lintian_brush.line_editor import LineEditor

resolution = ""


def update_configure_ac(path):
    global resolution
    if not os.path.exists(path):
        return

    with LineEditor(path, "b") as e:
        for lineno, line in e:
            m = re.fullmatch(
                b"\\s*AC_PATH_PROG\\s*"
                b"\\(\\s*(\\[)?(?P<variable>[A-Z_]+)(\\])?\\s*"
                b",\\s*(\\[)?pkg-config(\\])?\\s*"
                b"(,\\s*(\\[)?(?P<default>.*)(\\])?\\s*)?\\)\n",
                line,
            )

            if not m:
                continue

            issue = LintianIssue(
                "source",
                "autotools-pkg-config-macro-not-cross-compilation-" "safe",
                info=f"{name} (line {lineno})"
            )
            if not issue.should_fix():
                continue

            if m.group("variable") == b"PKG_CONFIG" and not m.group("default"):
                e[lineno] = b"PKG_PROG_PKG_CONFIG\n"
                resolution = (
                    "This patch changes it to use "
                    "PKG_PROG_PKG_CONFIG macro from pkg.m4."
                )
                # Build-Depend on pkg-config for pkg.m4
                with control:
                    control.source["Build-Depends"] = ensure_some_version(
                        control.source.get("Build-Depends", ""), "pkg-config"
                    )
            else:
                e[lineno] = line.replace(b"AC_PATH_PROG", b"AC_PATH_TOOL")
                resolution = "This patch changes it to use AC_PATH_TOOL."

            issue.report_fixed()


for name in ["configure.ac", "configure.in"]:
    update_configure_ac(name)


report_result(
    f"""Use cross-build compatible macro for finding pkg-config.

The package uses AC_PATH_PROG to discover the location of pkg-config(1). This
macro fails to select the correct version to support cross-compilation.

{resolution}

Refer to https://bugs.debian.org/884798 for details.
""",
    patch_name="ac-path-pkgconfig",
)



#!/usr/bin/python3

import sys
from contextlib import suppress

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import fixed_lintian_tag, report_result, warn

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)

valid_field_names = {
    "Files",
    "License",
    "Copyright",
    "Comment",
    "Upstream-Name",
    "Format",
    "Upstream-Contact",
    "Source",
    "Upstream",
    "Contact",
    "Name",
}

typo_fixed = set()
case_fixed = set()

with suppress(FileNotFoundError), Deb822Editor("debian/copyright") as updater:
    for paragraph in updater.paragraphs:
        for field in paragraph:
            if field in valid_field_names:
                continue
            if field.startswith("X-") and field[2:] in valid_field_names:
                if field[2:] in paragraph:
                    warn(f"Both {field} and {field[2:]} exist.")
                    continue
                value = paragraph[field]
                del paragraph[field]
                paragraph[field[2:]] = value
                typo_fixed.add((field, field[2:]))
                fixed_lintian_tag(
                    "source",
                    "field-name-typo-in-dep5-copyright",
                    f"{field} (line XX)",
                )
                continue

            for option in valid_field_names:
                if distance(field, option) == 1:
                    value = paragraph[field]
                    if option in paragraph and option.lower() != field.lower():
                        warn(
                            f"Found typo ({field} ⇒ {option}), "
                            f"but {option} already exists"
                        )
                        continue
                    del paragraph[field]
                    paragraph[option] = value
                    if option.lower() == field.lower():
                        case_fixed.add((field, option))
                    else:
                        typo_fixed.add((field, option))
                        fixed_lintian_tag(
                            "source",
                            "field-name-typo-in-dep5-copyright",
                            f"{field} (line XX)",
                        )
                    break


if case_fixed:
    kind = "case" + ("s" if len(case_fixed) > 1 else "")
else:
    kind = ""
if typo_fixed:
    if case_fixed:
        kind += " and "
    kind += "typo" + ("s" if len(typo_fixed) > 1 else "")

fixed_str = ", ".join(
    [
        f"{old} ⇒ {new}"
        for (old, new) in sorted(list(case_fixed) + list(typo_fixed))
    ]
)

report_result(f"Fix field name {kind} in debian/copyright ({fixed_str}).")



#!/usr/bin/python3

import os
import sys
from contextlib import suppress

from debmutate._rules import RulesEditor

from lintian_brush.fixer import LintianIssue, opinionated, report_result

certainty = None

if not opinionated() and os.listdir() == ["debian"]:
    # See https://salsa.debian.org/debian-ayatana-team/\
    # snapd-glib/-/merge_requests/6#note_358358
    sys.exit(0)

with suppress(FileNotFoundError), RulesEditor() as updater:
    for rule in updater.makefile.iter_rules(b"get-orig-source"):
        commands = rule.commands()
        if [b"uscan"] == [c.split(b" ")[0] for c in commands]:
            certainty = "certain"
        else:
            certainty = "possible"
        issue = LintianIssue(
            "source",
            "debian-rules-contains-unnecessary-get-orig-source-target",
            "",
        )
        if issue.should_fix():
            rule.clear()
            updater.makefile.drop_phony(b"get-orig-source")
            issue.report_fixed()


report_result(
    "Remove unnecessary get-orig-source-target.", certainty=certainty
)
